# -*- coding: utf-8 -*-

# @Time    : 2019/11/1 13:18
# @Email   : 986798607@qq.ele_ratio
# @Software: PyCharm
# @License: BSD 3-Clause

from abc import ABCMeta, abstractmethod

import numpy as np
from pymatgen.core.composition import Composition as PMGComp

from featurebox.featurizers.base import BaseFeaturizer
from featurebox.featurizers.extrastats import PropertyStats


class BaseCompositionFeaturizer(BaseFeaturizer, metaclass=ABCMeta):

    def __init__(self, *, elem_data, n_jobs=-1, on_errors='raise', return_type='any'):
        """
        Base class for composition feature.
        """

        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

        self._elements = elem_data
        self.__authors__ = ['TsumiNa']

    def featurize(self, comp):
        elems_, nums_ = [], []
        if isinstance(comp, PMGComp):
            comp = comp.as_dict()
        for e, n in comp.items():
            elems_.append(e)
            nums_.append(n)
        return self.mix_function(elems_, nums_)

    @abstractmethod
    def mix_function(self, elems, nums):
        """
        Parameters
        ----------
        elems: list
            Elements in compound.
        nums: list
            Number of each element.
        Returns
        -------
        descriptor: numpy.ndarray
        """


class WeightedAverage(BaseCompositionFeaturizer):
    def __init__(self, *, elem_data, n_jobs=-1, on_errors='raise', return_type='any'):
        """
        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input0 type.
            Default is ``any``
        """

        super().__init__(elem_data=elem_data, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, nums):
        elems_ = self._elements.loc[elems, :].values
        w_ = nums / np.sum(nums)
        return w_.dot(elems_)

    @property
    def feature_labels(self):
        return ['ave:' + s for s in self._elements]


class WeightedSum(BaseCompositionFeaturizer):
    def __init__(self, *, elem_data, n_jobs=-1, on_errors='raise', return_type='any'):
        """
        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input0 type.
            Default is ``any``
        """

        super().__init__(elem_data=elem_data, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, nums):
        elems_ = self._elements.loc[elems, :].values
        w_ = np.array(nums)
        return w_.dot(elems_)

    @property
    def feature_labels(self):
        return ['sum:' + s for s in self._elements]


class GeometricMean(BaseCompositionFeaturizer):
    def __init__(self, *, elem_data, n_jobs=-1, on_errors='raise', return_type='any'):
        """
        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input0 type.
            Default is ``any``
        """

        super().__init__(elem_data=elem_data, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, nums):
        elems_ = self._elements.loc[elems, :].values
        w_ = np.array(nums).reshape(-1, 1)
        tmp = elems_ ** w_
        return np.power(tmp.prod(axis=0), 1 / sum(w_))

    @property
    def feature_labels(self):
        return ['gmean:' + s for s in self._elements]


class HarmonicMean(BaseCompositionFeaturizer):
    def __init__(self, *, elem_data, n_jobs=-1, on_errors='raise', return_type='any'):
        """
        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input0 type.
            Default is ``any``
        """

        super().__init__(elem_data=elem_data, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, nums):
        elems_ = 1 / self._elements.loc[elems, :].values
        w_ = np.array(nums)
        tmp = w_.dot(elems_)

        return sum(w_) / tmp

    @property
    def feature_labels(self):
        return ['hmean:' + s for s in self._elements]


class WeightedVariance(BaseCompositionFeaturizer):
    def __init__(self, *, elem_data, n_jobs=-1, on_errors='raise', return_type='any'):
        """
        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input0 type.
            Default is ``any``
        """

        super().__init__(elem_data=elem_data, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, nums):
        elems_ = self._elements.loc[elems, :].values
        w_ = nums / np.sum(nums)
        mean_ = w_.dot(elems_)
        var_ = elems_ - mean_
        return w_.dot(var_ ** 2)

    @property
    def feature_labels(self):
        return ['var:' + s for s in self._elements]


class MaxPooling(BaseCompositionFeaturizer):
    def __init__(self, *, elem_data, n_jobs=-1, on_errors='raise', return_type='any'):
        """
        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input0 type.
            Default is ``any``
        """

        super().__init__(elem_data=elem_data, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, _):
        elems_ = self._elements.loc[elems, :]
        return elems_.max().values

    @property
    def feature_labels(self):
        return ['max:' + s for s in self._elements]


class MinPooling(BaseCompositionFeaturizer):
    def __init__(self, *, elem_data, n_jobs=-1, on_errors='raise', return_type='any'):
        """
        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input0 type.
            Default is ``any``
        """

        super().__init__(elem_data=elem_data, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, _):
        elems_ = self._elements.loc[elems, :]
        return elems_.min().values

    @property
    def feature_labels(self):
        return ['min:' + s for s in self._elements]


class ExtraMix(BaseCompositionFeaturizer):
    def __init__(self, *, elem_data, stats=("mean",), n_jobs=-1, on_errors='raise', return_type='any'):
        """
        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input0 type.
            Default is ``any``
        """

        super().__init__(elem_data=elem_data, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)
        self.stats = stats

    def mix_function(self, elems, nums):
        elems_ = self._elements.loc[elems, :].values

        all_attributes = []
        for stat in self.stats:
            all_attributes.append(PropertyStats.calc_stat(elems_, stat, nums))

        return np.array(all_attributes).ravel()

    @property
    def feature_labels(self):
        return ['%s:' % i + s for s in self._elements for i in self.stats]


class DepartElementFeaturizer(BaseCompositionFeaturizer):
    def __init__(self, *, elem_data, n_composition, n_jobs=-1, on_errors='raise', return_type='any'):
        """
        just for same number element for one conmpund.
        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input0 type.
            Default is ``any``
        """

        super().__init__(elem_data=elem_data, n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)
        self.n_composition = n_composition

    def mix_function(self, elems, nums=None):
        elems_ = self._elements.loc[elems, :].values
        return elems_.ravel(order='F')

    def featurize(self, comp):
        elems_, nums_ = [], []
        if isinstance(comp, PMGComp):
            comp = comp.to_reduced_dict
        for e, n in comp.items():
            elems_.append(e)
            nums_.append(n)
        return self.mix_function(elems_, nums_)

    @property
    def feature_labels(self, ):

        return [str(s) + "_" + str(n) for s in self._elements for n in range(self.n_composition)]



import copy

import numpy as np
import torch
import torch.nn.functional as F
from matplotlib import cm
from matplotlib.cm import ScalarMappable
from torchvision.transforms import Normalize as Nor


def _d255(x):
    return x / 255


def _m255(x):
    return x * 255


def _check255(x):
    if isinstance(x, torch.Tensor):
        return torch.max(x) > 1.1
    elif isinstance(x, np.ndarray):
        return np.max(x) > 1.1
    else:
        raise TypeError("x must be torch.Tensor or np.ndarray")


def trans_type(x, tp="float"):
    if isinstance(x, torch.Tensor):
        if tp == "int":
            f = x.int()
        else:
            f = x.float()
    elif isinstance(x, np.ndarray):
        if tp == "int":
            f = x.astype(np.int8)
        else:
            f = x.astype(np.float32)
    else:
        raise TypeError("x must be torch.Tensor or np.ndarray")

    return f


def tensor3c(tensor):
    return torch.is_tensor(tensor) and tensor.ndimension() >= 3


def ndarray3c(tensor):
    return isinstance(tensor, np.ndarray) and tensor.ndim >= 3


###################


def _mul255_(x):
    if _check255(x):
        pass
    else:
        x *= 255
    x = trans_type(x, tp="int")
    return x


def mul255(x):
    x1 = copy.deepcopy(x)
    return _mul255_(x1)


def _div255_(x):
    x = trans_type(x, tp="float")
    if _check255(x):
        x /= 255
    else:
        pass
    return x


def div255(x):
    x1 = copy.deepcopy(x)
    return _div255_(x1)


def trans_to_numpy(x, channel=True):
    x1 = copy.deepcopy(x)
    return _trans_to_numpy_(x1, channel=channel)


def _trans_to_numpy_(x, channel=True):
    # x = x.squeeze()

    if isinstance(x, torch.Tensor):
        if tensor3c(x) and channel:
            x = x.numpy()
            dim = list(range(x.ndim))
            dim.remove(0)
            dim.append(0)

            return x.transpose(dim)
        else:
            return x.numpy()
    elif isinstance(x, np.ndarray):
        return x
    else:
        raise TypeError("x must be torch.Tensor or np.ndarray")


def trans_to_tensor(x, channel=True):
    x1 = copy.deepcopy(x)
    return _trans_to_tensor_(x1, channel=channel)


def _trans_to_tensor_(x, channel=True):
    # x = x.squeeze()
    if isinstance(x, torch.Tensor):
        return x
    elif isinstance(x, np.ndarray):
        if ndarray3c(x) and channel:
            dim = list(range(x.ndim))
            dim = [-1] + dim
            dim.pop()
            return torch.from_numpy(x.transpose(dim))
        else:
            return torch.from_numpy(x)
    else:
        raise TypeError("x must be torch.Tensor or np.ndarray")


def trans_to_tensors(x, channel=True):
    return torch.stack([trans_to_tensor(xi, channel=channel) for xi in x], dim=0)


def trans_to_numpys(x, channel=True):
    return np.stack([trans_to_numpy(xi, channel=channel) for xi in x])


class Normalize(Nor):

    def __init__(self, mean, std, inplace=False):
        super().__init__(mean, std, inplace=inplace)

    def transform(self, tensor):
        return self.__call__(self, tensor)

    def inverse_transform(self, tensor):
        return self._it(tensor)

    def _it(self, tensor):
        mean = self.mean
        std = self.std
        inplace = self.inplace
        if not torch.is_tensor(tensor) and tensor.ndimension() == 3:
            raise TypeError('tensor is not a torch image.')

        if not inplace:
            tensor = tensor.clone()

        dtype = tensor.dtype
        mean = torch.as_tensor(mean, dtype=dtype, device=tensor.device)
        std = torch.as_tensor(std, dtype=dtype, device=tensor.device)
        tensor = torch.mul(tensor, std[:, None, None]) + mean[:, None, None]

        return tensor


normalize = Normalize(
    mean=[0.485, 0.456, 0.406],
    std=[0.229, 0.224, 0.225]
)


def visualize(img, cam, return_type="t"):
    """
    Synthesize an image with CAM to make a result image.
    Args:
        img: (Tensor) shape => (1, 3c, H, W)
        cam: (Tensor) shape => (1, 1, H', W')
    Return:
        synthesized image (Tensor): shape =>(1, 3, H, W)
    """
    img = div255(img.cpu())
    cam = div255(cam.cpu())

    if cam.dim() == 2:
        cam = cam.unsqueeze(dim=0)
    if cam.dim() == 3:
        cam = cam.unsqueeze(dim=0)
    if img.dim() == 2:
        img = img.unsqueeze(dim=0)
    if img.dim() == 3:
        img = img.unsqueeze(dim=0)

    *_, H, W = img.shape
    cam = F.interpolate(cam, size=(H, W), mode='bilinear', align_corners=False)

    img = img.squeeze(0).numpy()
    cam = cam.squeeze(0).numpy()
    cam = cam.transpose((1, 2, 0))
    img = img.transpose((1, 2, 0))
    cam = cam.astype(np.float32)
    cam = cam.squeeze()

    sc = ScalarMappable(cmap=cm.jet)
    cam = sc.to_rgba(cam)[:, :, :-1]
    camp = cam

    result = camp + img
    result = result / (result.max())
    if return_type is "t":
        result = trans_to_tensor(result)
    return result



import time

from featurebox.symbol.functions.dimfunc import dless, Dim
from featurebox.symbol.calculation.scores import compile_context
from numpy import random
from sklearn.datasets import load_boston

from featurebox.symbol.base import SymbolSet, SymbolTree

if __name__ == "__main__":
    # data
    data = load_boston()
    x = data["data"]
    y = data["target"]
    c = [6, 3, 4]
    # unit
    from sympy.physics.units import kg

    x_u = [kg] * 13
    y_u = kg
    c_u = [dless, dless, dless]

    x, x_dim = Dim.convert_x(x, x_u, target_units=None, unit_system="SI")
    y, y_dim = Dim.convert_xi(y, y_u)
    c, c_dim = Dim.convert_x(c, c_u)

    t = time.time()

    # symbolset
    pset0 = SymbolSet()
    pset0.add_features(x, y, x_dim=x_dim, y_dim=y_dim, x_group=[[1, 2], [3, 4, 5]])
    pset0.add_constants(c, c_dim=c_dim, c_prob=None)
    pset0.add_operations(power_categories=(2, 3, 0.5),
                         categories=("Add", "Mul", "Sub", "Div", "exp"),
                         self_categories=None)

    random.seed(0)
    z = time.time()
    sl = [SymbolTree.genGrow(pset0, 3, 4) for _ in range(100)]
    a = time.time()
    sl = [compile_context(sli, pset0.context, pset0.gro_ter_con) for sli in sl]
    b = time.time()

    print(b - a, a - z)



# -*- coding: utf-8 -*-

# @Time    : 2020/1/8 16:28
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause
import numpy as np


def get_rol(axis_all, axis):
    return [j for i, j in enumerate(axis_all) if i != axis]


def data_resize(x):
    return x


def rotatespilt(TDarray, angle=10, axis=0, axis_site=None, resize=False):
    """
    Rotate the np.array with an center and axis.

    Parameters
    ----------
    TDarray:np.ndarray
        three Dimension np.ndarray
    angle:int
        angle from 0 to 180
    axis:int
        the rotate axis
    axis_site:tuple
        rotate center
    resize:bool
        # todo
    Returns
    -------
    result:list of np.ndarray
        All 2d array with different angle.

    """
    assert TDarray.ndim == 3
    angle = np.pi / 180 * angle

    slices = []
    for j, i in enumerate(np.arange(0, np.pi, angle)):
        try:
            data_all = _rotatespilt(TDarray, angle=i, axis=axis, axis_site=axis_site)
            if resize:
                data_all = data_resize(data_all)
            slices.append(data_all)
        except UserWarning:
            print('except the angle {}'.format(i / np.pi * 180))
    return slices


def _rotatespilt(TDarray, angle=np.pi / 4, axis=0, axis_site=None):
    # 得到两轴长度
    if axis_site is None:
        axis_site = (0, 0)
    shape_rol = get_rol(TDarray.shape, axis)
    add_site = [int(i * j) for i, j in zip(axis_site, shape_rol)]
    add_site = np.array(add_site).reshape((-1, 1))
    if angle == np.pi / 2:
        if axis == 0:
            data = np.squeeze(TDarray[:, add_site[0], :])
        elif axis == 1:
            data = np.squeeze(TDarray[add_site[0], :, :])
        else:
            data = np.squeeze(TDarray[:, :, add_site[0]])
        data_all = data
    else:
        # 得到两轴
        # axis_all = list(range(TDarray.ndim))
        # axis_rol = get_rol(axis_all, axis)

        # 网格
        meshes = np.meshgrid(*[range(i) for i in shape_rol])
        meshes = np.array([_.ravel() for _ in meshes])

        # 距离
        s1 = meshes[0] - add_site[0]
        s2 = np.tan(angle) * (meshes[1] - add_site[1])
        distance = np.abs(s1 - s2)
        # show = np.array((s1, meshes[1] - add_site[1], s2, distance))
        # 间距
        site = np.where(distance < 2)[0]
        point = meshes[:, site]

        point_max = np.max(point, axis=1)
        point_min = np.min(point, axis=1)
        point_dis = point_max - point_min
        range_ = int(np.sqrt(np.sum((point_dis ** 2)))) + 1

        if angle < 1.57:
            start_point = point_min
            # end_point = point_max
        else:
            start_point = (point_min[0], point_max[1])
            # end_point = (point_max[0], point_min[1])

        dis = point - np.array(start_point).reshape(-1, 1)
        dis = np.sqrt(np.sum((dis ** 2), axis=0))
        point_near = []
        for i in range(range_):
            tem = np.abs(dis - i)
            if len(tem) <= 3:
                raise UserWarning(
                    "the slice is at a corner,with little point,please change the axis_site or the angle ")
            else:
                near = np.argpartition(tem, 3)[:3]
            point_near.append(point[:, near])

        data_all = []
        for point_near_i in point_near:
            if axis == 0:
                data = [TDarray[:, i, j] for i, j in point_near_i.T]
            elif axis == 1:
                data = [TDarray[i, :, j] for i, j in point_near_i.T]
            else:
                data = [TDarray[i, j, :] for i, j in point_near_i.T]

            data = np.mean(np.array(data), axis=0)
            data_all.append(data)
        data_all = np.array(data_all).T

    return data_all


# a = np.array(
#     [[[1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 1, 1, 1, 1]], [[0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], ],
#      [[3, 3, 2, 3, 3], [3, 3, 2, 3, 3], [3, 3, 2, 3, 3]]])
a = np.zeros((11, 12, 13))
s = rotatespilt(a, angle=10, axis=0, axis_site=(0.5, 0.5))



"""
Notes: the translation process
    the three function should be the same key.
    1.
    sym_vector_map: repr of SymbolTree to sympy.Function
    sym_dispose_map: repr of SymbolTree to "group" sympy.Function.
    2.
    np_map(): repr of "group" sympy.Function to numpy function
    3.
    dim_map(): repr of "group" sympy.Function to Dim function
    4.
    gsym_map():repr of "group" sympy.Function to universal sympy.Function
"""

__all__=["dimfunc","gsymfunc","newfunc","npfunc","symfunc"]




import unittest
import numpy as np

from featurebox.symbol.probability.preference import PreMap


class MyTestCase(unittest.TestCase):
    def setUp(self):
        self.premap = PreMap.from_shape(14)

    def test_down_self(self):
        self.premap.down_other_point(*[0, 1, 0.9])
        # self.premap.down_others(*[0, 5, 0.4])
        sums1 = np.sum(self.premap, axis=0)
        sums2 = np.sum(self.premap, axis=1)
        print(sums1)
        print(sums2)
        # [self.assertEqual(str(round(sumsi,3)), str(0.895)) for sumsi in sums2]



import numpy as np
from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.utils import check_array


def ger_magnitude(a):
    c = 0
    if abs(a) > 1:
        while a >= 1:
            a /= 10
            c += 1
    elif 1 >= abs(a) > 0:
        while a <= 1:
            a *= 10
            c += 1
        c = -c

    return c


def _scale(a):
    return 10 ** ger_magnitude(a)


class MagnitudeTransformer(TransformerMixin, BaseEstimator):
    """
    Transform x, y or c to near to 1, and store the transform Magnitude.
    """
    def __init__(self, standard=1, tolerate=0):
        self.standard = standard
        self.scale = None
        self.scale_y = None
        self.scale_c = None
        self.tolerate = tolerate

    def fit(self, X, y=None, group=2, apply=None, keep=None):
        """

        Parameters
        ----------
        X: np.ndarray
        y: np.ndarray
        group:group index of x
        apply: specific which index of x to transform
        keep: specific which index of x to not transform

        Returns
        -------

        """

        X = X.astype(np.float32)
        X = check_array(X, ensure_2d=True)
        n = X.shape[1]
        assert isinstance(X, np.ndarray)
        means = np.mean(X, axis=0)

        if not group:
            pass
        else:
            if isinstance(group, int):
                assert n > group > 1, "the len of group should in (2,x.shape[1]]"
                indexes = [_ for _ in range(n)]
                group = [indexes[i:i + group] for i in range(0, len(indexes), group)]
            for g in group:
                if len(g) > 0:
                    means[g] = np.mean(means[g])

        scale = np.array([_scale(i) for i in means])

        if self.tolerate:
            scale = np.array([i if not 10 ** (-self.tolerate) <= i <= 10 ** self.tolerate else 1 for i in scale])
        scale = scale.astype(np.float32)
        scale /= self.standard

        if apply is not None:
            if isinstance(apply, int):
                apply = [apply, ]
            li = list(range(n))
            keep = list(set(li) - set(apply))
            keep.sort()

        if keep is not None:
            if isinstance(keep, int):
                keep = [keep, ]
            for i in keep:
                scale[i] = 1

        self.scale = scale

        if y is not None:
            y = y.astype(np.float32)
            y = check_array(y, ensure_2d=False)
            assert isinstance(y, np.ndarray)
            means = np.mean(y)
            scale = _scale(means)
            if self.tolerate:
                scale = scale if not 10 ** (-self.tolerate) <= scale <= 10 ** self.tolerate else 1
            scale /= self.standard
            self.scale_y = scale
        return self

    def transform(self, X):
        if self.scale is None:
            raise NotImplementedError("the method should be fitted first.")
        else:
            return X / self.scale

    def inverse_transform(self, X):
        if self.scale_y is None:
            raise NotImplementedError("the method should be fitted first.")
        else:
            return self.scale_y * X

    def transform_y(self, y):
        if self.scale_y is None:
            raise NotImplementedError("the method should be fitted y first.")
        else:
            return y / self.scale_y

    def inverse_transform_y(self, y):
        if self.scale_y is None:
            raise NotImplementedError("the method should be fitted y first.")
        else:
            return self.scale_y * y

    def fit_transform_all(self, X, y, **fit_params):
        if y is None:
            # fit method of arity 1 (unsupervised transformation)
            return self.fit(X, **fit_params).transform(X)
        else:
            # fit method of arity 2 (supervised transformation)
            self.fit(X, y, **fit_params)
            return self.transform(X), self.transform_y(y)

    def fit_constant(self, c):
        if isinstance(c, float):
            c = [c, ]
        scale_c = []
        for ci in c:
            scale = _scale(ci)
            if self.tolerate:
                scale = scale if not 10 ** (-self.tolerate) <= scale <= 10 ** self.tolerate else 1
            scale /= self.standard
            scale_c.append(float(scale))
        scale_c = np.array(scale_c)
        scale_c.astype(np.float32)
        self.scale_c = scale_c
        return self

    def transform_constant(self, c):
        if self.scale_y is None:
            raise NotImplementedError("the method should be fitted c first.")
        if isinstance(c, float):
            c = [c, ]
        c = np.array([float(i) for i in c])
        c.astype(np.float32)
        return c / self.scale_c

    def inverse_transform_constant(self, c):
        if self.scale_y is None:
            raise NotImplementedError("the method should be fitted c first.")
        if isinstance(c, float):
            c = [c, ]
        c = np.array(c)
        c.astype(np.float32)
        return c * self.scale_c

    def fit_transform_constant(self, c):
        return self.fit_constant(c).transform_constant(c)



from collections import deque

import numpy as np
from deap import base
from deap import creator
from deap import tools
from deap.algorithms import varAnd
from deap.tools import mutShuffleIndexes
from sklearn.datasets import load_boston
from sklearn.svm import SVR


def eaSimple(population, toolbox, cxpb, mutpb, ngen, stats=None,
             halloffame=None, verbose=__debug__):
    """This algorithm reproduce the simplest evolutionary algorithm as
    presented in chapter 7 of [Back2000]_.

    :param population: A list of individuals.
    :param toolbox: A :class:`~deap.base.Toolbox` that contains the evolution
                    operators.
    :param cxpb: The probability of mating two individuals.
    :param mutpb: The probability of mutating an individual.
    :param ngen: The number of generation.
    :param stats: A :class:`~deap.tools.Statistics` object that is updated
                  inplace, optional.
    :param halloffame: A :class:`~deap.tools.HallOfFame` object that will
                       contain the best individuals, optional.
    :param verbose: Whether or not to log the statistics.
    :returns: The final population
    :returns: A class:`~deap.tools.Logbook` with the statistics of the
              evolution

    """
    logbook = tools.Logbook()
    logbook.header = ['gen', 'nevals'] + (stats.fields if stats else [])

    # Evaluate the individuals with an invalid fitness
    invalid_ind = [ind for ind in population if not ind.fitness.valid]
    fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
    for ind, fit in zip(invalid_ind, fitnesses):
        ind.fitness.values = fit

    if halloffame is not None:
        halloffame.update(population)

    best = deque([halloffame.items[0].fitness.values[0]], maxlen=15)

    # Begin the generational process
    for gen in range(1, ngen + 1):

        # Append the current generation statistics to the logbook
        record = stats.compile_(population + halloffame.items) if stats else {}
        logbook.record(gen=gen, nevals=len(invalid_ind), **record)
        if verbose:
            print(logbook.stream)

        # Select the next generation individuals
        offspring = toolbox.select(population, len(population) - halloffame.maxsize)
        offspring.extend(halloffame.items)
        # Vary the pool of individuals
        offspring = varAnd(offspring, toolbox, cxpb, mutpb)

        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

        # Update the hall of fame with the generated individuals
        if halloffame is not None:
            halloffame.update(offspring)

        best.append(halloffame.items[0].fitness.values[0])
        if sum(best) / best.maxlen == best[-1]:
            break
        # Replace the current population by the offspring
        population[:] = offspring

    return population, logbook


def generate_xi():
    return np.random.randint(0, 2)


def generate(space):
    return [generate_xi() for i in range(space)]


def GA(x, y, fit_func, pop_n=1000, hof_n=1, cxpb=0.6, mutpb=0.3, ngen=40, max_or_min="max", mut_indpb=0.05):
    x_space = x.shape[1]
    if max_or_min == "max":
        creator.create("FitnessMax", base.Fitness, weights=(1.0,))
    else:
        creator.create("FitnessMax", base.Fitness, weights=(-1.0,))
    creator.create("Individual", list, fitness=creator.FitnessMax)

    toolbox = base.Toolbox()
    # Attribute generator
    toolbox.register("generate_x", generate, x_space)
    toolbox.register("generate_xi", generate_xi, x_space)
    # Structure initializers
    toolbox.register("individual", tools.initIterate, creator.Individual,
                     toolbox.generate_x)
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)

    toolbox.register("evaluate", fit_func, x=x, y=y)
    toolbox.register("mate", tools.cxTwoPoint)
    toolbox.register("mutate", mutShuffleIndexes, indpb=mut_indpb)
    toolbox.register("select", tools.selTournament, tournsize=3)

    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("min", np.min)
    stats.register("max", np.max)

    pop = toolbox.population(n=pop_n)
    hof = tools.HallOfFame(hof_n)

    eaSimple(pop, toolbox, cxpb=cxpb, mutpb=mutpb, ngen=ngen,
             stats=stats, halloffame=hof, verbose=True)
    for i in hof.items:
        print(i, i.fitness)
    return hof.items


if __name__ == "__main__":
    data = load_boston()
    x = data.data
    y = data.target
    svr = SVR(gamma="scale")


    def fitn(ind, x, y):
        index = np.where(np.array(ind) == 1)[0]
        x = x[:, index]
        if x.shape[1] > 1:

            svr = SVR(gamma="scale")
            svr.fit(x, y)
            sc = svr.score(x, y)

            return sc,
        else:
            return 0,


    best = GA(x, y, fitn, pop_n=500, hof_n=1, cxpb=0.8, mutpb=0.4, ngen=10, max_or_min="max", mut_indpb=0.1)



# -*- coding: utf-8 -*-

# @Time   : 2019/5/23 20:40
# @Author : Administrator
# @Project : feature_preparation
# @FileName: backforward.py
# @Software: PyCharm

"""forward_and_back feature elimination for feature ranking"""

import copy
from functools import partial

import numpy as np
from sklearn.base import BaseEstimator, is_classifier
from sklearn.base import MetaEstimatorMixin
from sklearn.base import clone
from sklearn.feature_selection._base import SelectorMixin
from sklearn.metrics import check_scoring
from sklearn.model_selection import check_cv
from sklearn.model_selection._validation import _score, cross_val_score
from sklearn.utils.metaestimators import if_delegate_has_method, _safe_split
from sklearn.utils.validation import check_is_fitted, check_X_y, check_random_state

from featurebox.tools.tool import parallelize
from .mutibase import MutiBase


def _baf_single_fit(train, test, baf, estimator, X, y, scorer, random_state):
    """"""
    X_train, y_train = _safe_split(estimator, X, y, train)
    X_test, y_test = _safe_split(estimator, X, y, test, train)
    baf_i = clone(baf)
    baf_i.random_state = random_state
    baf_i._fit(X_train, y_train)
    return baf_i.support_, _score(baf_i.estimator_, baf_i.transform(X_test, ), y_test, scorer), baf_i.score_


class BackForward(BaseEstimator, MetaEstimatorMixin, SelectorMixin, MutiBase):
    """

    Parameters
    ----------
    estimator :
        A supervised learning estimator with a ``_fit`` method that provides
        information about feature importance either through a ``coef_``
        attribute or through a ``feature_importances_`` attribute.

    n_type_feature_to_select : int or None (default=None)
        The number of feature to selection. If `None`, selection the features with best score .


    Attributes
    ----------
    n_feature_ : int
        The number of selected feature.

    support_ : array of shape [n_feature]
        The mask of selected feature.

    estimator_ : object
        The external estimator _fit on the reduced dataset.

    """

    def __init__(self, estimator, n_type_feature_to_select=None, primary_feature=None, muti_grade=2, muti_index=None,
                 must_index=None, tolerant=0.01, verbose=0, random_state=None):
        super().__init__(muti_grade=muti_grade, muti_index=muti_index, must_index=must_index)
        self.estimator = estimator
        self.n_type_feature_to_select = n_type_feature_to_select
        self.primary_feature = primary_feature
        self.verbose = verbose
        self.score_ = []
        self.random_state = random_state
        self.tolerant = tolerant

    @property
    def _estimator_type(self):
        return self.estimator._estimator_type

    def fit(self, X, y):
        """Fit the baf model and then the underlying estimator on the selected
           feature.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_feature]
            The training input0 samples.

        y : array-like, shape = [n_samples]
            The target values.
        """
        return self._fit(X, y)

    def _fit(self, x, y):

        def add_slice(slice10, slice20):
            best0 = score(slices=slice10)

            for add1 in list(slice20):
                # or random.choice() >= score(slice1):
                slice10.append(add1)
                test = score(slices=slice10)
                slice1_, test_ = sub_slice(slice10)
                if test_ >= max(best0, test):
                    best0 = test_
                    slice10 = slice1_
                elif test > max(best0, test_) + 2 * self.tolerant:
                    best0 = test
                else:
                    slice10.remove(add1)
                if self.verbose > 0:
                    print("Fitting estimator with {} feature {}".format(len(slice10), best0))
                self.score_.append((tuple(slice10), best0))
            return slice10

        def sub_slice(slice10):
            best0 = score(slices=slice10)
            slice0 = list(copy.deepcopy(slice10))
            slice10 = list(slice10)
            ran.shuffle(slice10)
            if self.check_must:
                slice10 = [_ for _ in slice10 if _ not in self.check_must]
            for sub in list(slice10):
                slice0.remove(sub)
                test0 = score(slices=slice0)
                if test0 > best0 - self.tolerant:
                    best0 = test0
                else:
                    slice0.append(sub)
                # print(slice0, best)
            return slice0, best0

        def sub_slice_force(slice10):
            best0 = score(slices=slice10)
            slice0 = list(copy.deepcopy(slice10))
            ran.shuffle(slice0)

            while len(slice0) > self.n_type_feature_to_select:
                slice_all = []
                slice10 = list(copy.deepcopy(slice0))
                if self.check_must:
                    slice10 = [_ for _ in slice10 if _ not in self.check_must]
                le = len(slice10)
                while le:
                    slice_all.append(copy.deepcopy(slice0))
                    le -= 1
                for slice_, sub in zip(slice_all, slice10):
                    slice_.remove(sub)
                test0 = [score(slices=i) for i in slice_all]
                index = np.argmax(test0)
                best0 = max(test0)
                slice0 = slice_all[index]
                # print(slice0, best0)
                if self.verbose > 0:
                    print("Fitting estimator with {} feature {}".format(len(slice0), best0))
                self.score_.append((tuple(slice0), best0))
            return slice0, best0

        def score_pri(slices, x0, y0):
            slices = list(slices)
            if len(slices) <= 1:
                score0 = - np.inf
            else:
                slices = self.feature_unfold(slices)
                data_x0 = x0[:, slices]

                self.estimator.fit(data_x0, y0)
                if hasattr(self.estimator, 'best_score_'):
                    score0 = np.mean(self.estimator.best_score_)
                else:
                    score0 = np.mean(cross_val_score(self.estimator, data_x0, y0, cv=5))
            return score0

        score = partial(score_pri, x0=x, y0=y)

        self.score_ = []
        x, y = check_X_y(x, y, "csc")
        assert all((self.check_must, self.check_muti)) in [True, False]
        # Initialization
        if self.check_muti:
            n_feature = (self.muti_index[1] - self.muti_index[0]) // self.muti_grade + self.muti_index[0]
        else:
            n_feature = x.shape[1]
        if self.primary_feature is None:
            primary_feature = n_feature // 2
        else:
            primary_feature = self.primary_feature

        feature_list = list(range(x.shape[1]))
        fold_feature_list = self.feature_fold(feature_list)

        ran = check_random_state(self.random_state)
        slice1 = ran.choice(fold_feature_list, primary_feature, replace=False)
        slice1 = list(self.feature_must_fold(slice1))
        slice2 = list(set(fold_feature_list) - set(slice1))
        ran.shuffle(slice2)

        slice1 = add_slice(slice1, slice2)
        if isinstance(self.n_type_feature_to_select, int) and len(slice1) > self.n_type_feature_to_select:
            slice1, best = sub_slice_force(slice1)

        select_feature = self.feature_unfold(slice1)
        su = np.zeros(x.shape[1], dtype=np.bool)
        su[select_feature] = 1
        self.support_ = su
        self.estimator_ = clone(self.estimator)
        self.estimator_.fit(x[:, select_feature], y)
        self.n_feature_ = len(select_feature)

        return self

    @if_delegate_has_method(delegate='estimator')
    def predict(self, X):
        """Reduce X to the selected feature and then Fit using the
           underlying estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        Returns
        -------
        y : array of shape [n_samples]
            The predicted target values.
        """
        check_is_fitted(self, 'estimator_')
        return self.estimator_.predict(self.transform(X))

    @if_delegate_has_method(delegate='estimator')
    def score(self, X, y):
        """Reduce X to the selected feature and then return the score of the
           underlying estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        y : array of shape [n_samples]
            The target values.
        """
        check_is_fitted(self, 'estimator_')
        return self.estimator_.score(self.transform(X), y)

    def _get_support_mask(self):
        check_is_fitted(self, 'support_')
        return self.support_


class BackForwardCV(MetaEstimatorMixin, SelectorMixin):
    """Feature ranking with recursive feature elimination and cross-validated
    selection of the best number of feature.

    See glossary entry for :term:`cross-validation estimator`.

    Read more in the :ref:`User Guide <baf>`.

    Parameters
    ----------
    estimator : object
        A supervised learning estimator with a ``_fit`` method that provides
        information about feature importance either through a ``coef_``
        attribute or through a ``feature_importances_`` attribute.


    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 3-fold cross-validation,
        - integer, to specify the number of folds.
        - :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`sklearn.model_selection.StratifiedKFold` is used. If the
        estimator is a classifier or if ``y`` is neither binary nor multiclass,
        :class:`sklearn.model_selection.KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.

        .. versionchanged:: 0.20
            ``cv`` default value of None will change from 3-fold to 5-fold
            in v0.22.

    scoring : string, callable or None, optional, (default=None)
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        ``scorer(estimator, X, y)``.

    verbose : int, (default=0)
        Controls verbosity of output.

    n_jobs : int or None, optional (default=None)
        Number of cores to run in parallel while fitting across folds.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    Attributes
    ----------
    n_feature_ : int
        The number of selected feature with cross-validation.

    support_ : array of shape [n_feature]
        The mask of selected feature.

    estimator_ : object
        The external estimator _fit on the reduced dataset.
    """

    def __init__(self, estimator, n_type_feature_to_select=None, primary_feature=None, muti_grade=2, muti_index=None,
                 must_index=None, verbose=0, random_state=None,
                 cv=5, scoring="r2", n_jobs=None):
        self.estimator = estimator
        self.cv = cv
        self.scoring = scoring
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.n_type_feature_to_select = n_type_feature_to_select
        self.primary_feature = primary_feature
        self.muti_grade = muti_grade
        self.muti_index = muti_index
        self.must_index = must_index
        self.score_ = []
        self.random_state = random_state
        # spath.support_cv = ""
        # spath.score_cv = ""
        # spath.support_ = ""
        # spath.score_ = ""
        # spath.estimator_ = ""
        # spath.n_feature_ = ""

    def fit(self, X, y, groups=None):
        """Fit the baf model and automatically tune the number of selected
           feature.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_feature]
            Training vector, where `n_samples` is the number of samples and
            `n_feature` is the total number of feature.

        y : array-like, shape = [n_samples]
            Target values (integers for classification, real numbers for
            regression).

        groups : array-like, shape = [n_samples], optional
            cal_group labels for the samples used while splitting the dataset into
            train/test set.
        """
        X, y = check_X_y(X, y, "csr")
        # Initialization
        cv = check_cv(self.cv, y, is_classifier(self.estimator))
        scorer = check_scoring(self.estimator, scoring=self.scoring)
        ran = check_random_state(self.random_state)

        baf = BackForward(estimator=self.estimator,
                          n_type_feature_to_select=self.n_type_feature_to_select,
                          verbose=self.verbose, primary_feature=self.primary_feature,
                          muti_grade=self.muti_grade, muti_index=self.muti_index,
                          must_index=self.must_index, random_state=ran)

        func = partial(_baf_single_fit, baf=baf, estimator=self.estimator, X=X, y=y, scorer=scorer, random_state=ran)

        scores = parallelize(n_jobs=self.n_jobs, func=func, iterable=cv.split(X, y, groups), respective=True)

        support, scores, score_step = zip(*scores)
        best_support = support[np.argmax(scores)]
        best_score = max(scores)
        # Re-execute an elimination with best_k over the whole set

        # Set final attributes
        self.support_step = score_step
        self.support_cv = support
        self.support_ = best_support
        self.score_cv = scores
        self.score_ = best_score
        self.estimator_ = clone(self.estimator)
        self.estimator_.fit(X[:, self.support_], y)
        self.n_feature_ = np.count_nonzero(support)
        return self

    @if_delegate_has_method(delegate='estimator')
    def predict(self, X):
        """Reduce X to the selected feature and then Fit using the
           underlying estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        Returns
        -------
        y : array of shape [n_samples]
            The predicted target values.
        """
        check_is_fitted(self, 'estimator_')
        return self.estimator_.predict(self.transform(X))

    @if_delegate_has_method(delegate='estimator')
    def score(self, X, y):
        """Reduce X to the selected feature and then return the score of the
           underlying estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        y : array of shape [n_samples]
            The target values.
        """
        check_is_fitted(self, 'estimator_')
        return self.estimator_.score(self.transform(X), y)

    def _get_support_mask(self):
        check_is_fitted(self, 'support_')
        return self.support_



# -*- coding: utf-8 -*-

# @Time    : 2019/11/11 13:15
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause
import functools

import sympy
from deap.gp import PrimitiveTree, PrimitiveSet


class ExpressionTree(PrimitiveTree):
    """

    """
    hasher = str

    def __init__(self, content):
        super(ExpressionTree, self).__init__(content)

    def __repr__(self):
        """Symbolic representation of the expression tree."""
        repr1 = ''
        stack = []
        for node in self:
            stack.append((node, []))
            while len(stack[-1][1]) == stack[-1][0].arity:
                prim, args = stack.pop()
                repr1 = prim.format(*args)
                if len(stack) == 0:
                    break
                stack[-1][1].append(repr1)
        return repr1

    def __hash__(self):
        return hash(self.hasher(self))

    def __eq__(self, other):
        return hash(self) == hash(other)

    @property
    def terminals(self):
        """Return terminals that occur in the expression tree."""
        return [primitive for primitive in self if primitive.arity == 0]

    @property
    def pri_site(self):
        return [i for i, primitive in enumerate(self) if primitive.arity >= 1]

    @property
    def ter_site(self):
        return [i for i, primitive in enumerate(self) if primitive.arity == 0]

    @property
    def primitives(self):
        """Return primitives that occur in the expression tree."""
        return [primitive for primitive in self if primitive.arity >= 1]


class ExpressionSet(PrimitiveSet):

    def __init__(self, set_name, prefix="x"):
        super(ExpressionSet, self).__init__(set_name, 0, prefix=prefix)
        self.dimtext = {"__builtins__": None}

    def addTerminal(self, terminal, name=None, dim=None):
        PrimitiveSet.addTerminal(self, terminal, name=name)
        if name is None:
            name = terminal.__name__
        self.dimtext[name] = dim


def ExpressionSetFill(x_name, rep_name=None, power_categories=None, categories=("Add", "Mul", "Self", "exp"),
                      partial_categories=None, self_categories=None, dim=None):
    """

    Parameters
    ----------
        :type partial_categories: double list
        partial_categories = [["Add","Mul"],["x4"]]
        :type power_categories: list
        index_categories=[0.5,1,2,3]
        :type dim: list,tuple
        :type x_name: list,tuple
        :type rep_name: list,tuple
        :type categories: list,tuple
        :param self_categories:
        def rem(a):
            return 1-a
        self_categories = [[rem, 1, 'rem']]
    """

    def Div(left, right):
        return left / right

    def Sub(left, right):
        return left - right

    def zeroo(_):
        return 0

    def oneo(_):
        return 1

    def remo(a):
        return 1 - a

    def self(_):
        return _

    functions2 = {"Add": sympy.Add, 'Sub': Sub, 'Mul': sympy.Mul, 'Div': Div}
    functions1 = {"sin": sympy.sin, 'cos': sympy.cos, 'exp': sympy.exp, 'log': sympy.ln,
                  'Abs': sympy.Abs, "Neg": functools.partial(sympy.Mul, -1.0),
                  "Rec": functools.partial(sympy.Pow, e=-1.0),
                  'Zeroo': zeroo, "Oneo": oneo, "Remo": remo, "Self": self}

    pset0 = ExpressionSet('main')

    if power_categories:
        for j, i in enumerate(power_categories):
            pset0.addPrimitive(functools.partial(sympy.Pow, e=i), arity=1, name='pow%s' % j)

    for i in categories:
        if i in functions1:
            pset0.addPrimitive(functions1[i], arity=1, name=i)
        if i in functions2:
            pset0.addPrimitive(functions2[i], arity=2, name=i)

    if partial_categories:
        for partial_categoriesi in partial_categories:
            for i in partial_categoriesi[0]:
                for j in partial_categoriesi[1]:
                    if i in ["Mul", "Add"]:
                        pset0.addPrimitive(functools.partial(functions2[i], sympy.Symbol(j)), arity=1,
                                           name="{}_{}".format(i, j))
                    elif i in ["Div", "Sub"]:
                        pset0.addPrimitive(functools.partial(functions2[i], right=sympy.Symbol(j)), arity=1,
                                           name="{}_{}".format(i, j))
                    else:
                        pass
    if self_categories:
        for i in self_categories:
            pset0.addPrimitive(i[0], i[1], i[2])

    # define terminal
    if not rep_name:
        rep_name = ['{}{}'.format("x", i) for i, _ in enumerate(x_name)]

    rep_name = [sympy.Symbol(i) for i in rep_name]

    if dim is None:
        dim = [None] * len(rep_name)

    assert len(dim) == len(x_name) == len(rep_name)

    for sym, dimi in zip(rep_name, dim):
        pset0.addTerminal(sym, name=str(sym), dim=dimi)

    pset0.rep_name_list = rep_name
    pset0.real_name_list = x_name
    pset0.dim_list = dim

    return pset0



from sklearn.base import BaseEstimator, TransformerMixin, MultiOutputMixin
from sklearn.metrics import check_scoring

from featurebox.symbol import flow
from featurebox.symbol.base import SymbolSet
from featurebox.symbol.calculation.scores import calculate_y_unpack
from featurebox.symbol.calculation.translate import general_expr
from featurebox.symbol.flow import MutilMutateLoop
from featurebox.symbol.functions.dimfunc import Dim


class SymbolLearning(BaseEstimator, MultiOutputMixin, TransformerMixin):
    """One simplify API for flow.\n
    The detailed functions should turn to the loop of featurebox.symbol.flow.\n
    The SymbolLearning is time costing and are not suit for GridSearchCV, the cross_validate are builtin """

    def __str__(self):
        return str(self.loop)

    def __init__(self, *args, loop=None, **kwargs):
        """
        Parameters
        ----------
        pset:SymbolSet
            the feature x and traget y and others should have been added.
        loop: str
            featurebox.symbol.flow.BaseLoop
            [“BaseLoop”,”MutilMutateLoop“,“OnePointMutateLoop”, ”DimForceLoop“...]
        pop:int
            number of popolation
        gen:int
            number of generation
        mutate_prob:float
            probability of mutate
        mate_prob:float
            probability of mate(crossover)
        initial_max:int
            max initial size of expression when first producing.
        initial_min : None,int
            max initial size of expression when first producing.
        max_value:int
            max size of expression
        hall:int,>=1
            number of HallOfFame(elite) to maintain
        re_hall:None or int>=2
            Notes: only vaild when hall
            number of HallOfFame to add to next generation.
        re_Tree: int
            number of new features to add to next generation.
            0 is false to add.
        personal_map:bool or "auto"
            "auto" is using premap and with auto refresh the premap with individual.\n
            True is just using constant premap.\n
            False is just use the prob of terminals.
        scoring: list of Callbale, default is [sklearn.metrics.r2_score,]
            See Also sklearn.metrics
        score_pen: tuple of  1, -1 or float but 0.
            >0 : max problem, best is positive, worse -np.inf
            <0 : min problem, best is negative, worse np.inf
            Notes:
            if multiply score method, the scores must be turn to same dimension in preprocessing
            or weight by score_pen. Because the all the selection are stand on the mean(w_i*score_i)
            Examples: [r2_score] is [1],
        cv:sklearn.model_selection._split._BaseKFold,int
            the shuffler must be False,
            default=1 means no cv
        filter_warning:bool
            filter warning or not
        add_coef:bool
            add coef in expression or not.
        inter_add：bool
            add intercept constant or not
        inner_add:bool
            dd inner coeffcients or not
        n_jobs:int
            default 1, advise 6
        batch_size:int
            default 40, depend of machine
        random_state:int
            None,int
        cal_dim:bool
            excape the dim calculation
        dim_type:Dim or None or list of Dim
            "coef": af(x)+b. a,b have dimension,f(x) is not dnan. \n
            "integer": af(x)+b. f(x) is interger dimension. \n
            [Dim1,Dim2]: f(x) in list. \n
            Dim: f(x) ~= Dim. (see fuzzy) \n
            Dim: f(x) == Dim. \n
            None: f(x) == pset.y_dim
        fuzzy:bool
            choose the dim with same base with dim_type,such as m,m^2,m^3.
        stats:dict
            details of logbook to show. \n
            Map:\n
            values
                = {"max": np.max, "mean": np.mean, "min": np.mean, "std": np.std, "sum": np.sum}
            keys
                = {\n
                   "fitness": just see fitness[0], \n
                   "fitness_dim_max": max problem, see fitness with demand dim,\n
                   "fitness_dim_min": min problem, see fitness with demand dim,\n
                   "dim_is_target": demand dim,\n
                   "coef":  dim is true, coef have dim, \n
                   "integer":  dim is integer, \n
                   ...
                   }
            if stats is None, default is :\n
                stats = {"fitness_dim_max": ("max",), "dim_is_target": ("sum",)}   for cal_dim=True
                stats = {"fitness": ("max",)}                                      for cal_dim=False
            if self-definition, the key is func to get attribute of each ind./n
            Examples:
                def func(ind):\n
                    return ind.fitness[0]
                stats = {func: ("mean",), "dim_is_target": ("sum",)}
        verbose:bool
            print verbose logbook or not
        tq:bool
            print progress bar or not
        store:bool or path
            bool or path
        stop_condition:callable
            stop condition on the best ind of hall, which return bool,the true means stop loop.
            Examples:
                def func(ind):\n
                    c = ind.fitness.values[0]>=0.90
                    return c
        """
        self.args = args
        self.kwargs = kwargs
        if loop is None:
            loop = MutilMutateLoop
        if isinstance(loop, str):
            loop = getattr(flow, loop)

        self.loop = loop

    def fit(self, X=None, y=None, c=None, x_group=None, pset=None):
        """

        If pset is None, one simple pset are generate with no dimension calculation, But just with x_group.\n
        If need more self-definition, use one defined SymbolSet object to pset.\n
        Examples:
            pset = SymbolSet()\n
            pset.add_features_and_constants(...)\n
            pset.add_operations(...)\n
            ...\n
            ...SymbolLearning().fit(pset=pset)\n

        Parameters
        ----------
        X:np.ndarray
        
        y:np.ndarray
        
        c:list of float
        
        x_group:list of list
            Group of x.\n
            See Also pset.add_features_and_constants
        pset:SymbolSet
            See Also SymbolSet

        """
        if pset is None:
            pset = SymbolSet()
            pset.add_features_and_constants(X, y, c, x_dim=1, y_dim=1, c_dim=1, x_prob=None,
                                            c_prob=None, x_group=x_group, feature_name=None)
            pset.add_operations(power_categories=(2, 3, 0.5),
                                categories=("Add", "Mul", "Sub", "Div"))

        self.loop = self.loop(pset, *self.args, **self.kwargs)
        hall = self.loop.run()
        self.best_one = hall.items[0]
        try:
            expr = general_expr(self.best_one.coef_expr, self.loop.cpset)
            self.expr_type = "single"
        except (RecursionError, RuntimeWarning):
            expr = self.best_one.coef_expr
            self.expr_type = "group"

        self.expr = expr
        self.y_dim = self.best_one.y_dim
        self.fitness = self.best_one.fitness.values[0]

    def _predict_by_single(self, X):

        terminals = self.loop.cpset.init_free_symbol
        indexs = [int(i.name.replace("x", "")) for i in terminals if "x" in i.name]
        X = [xi for xi in X.T]
        X = [X[indexi] for indexi in indexs]

        c = []
        for i in self.loop.cpset.data_x_dict.keys():
            if "c" in i:
                c.append(self.loop.cpset.data_x_dict[i])
        X_and_c = X + c
        pre_y = calculate_y_unpack(self.expr, X_and_c, terminals)
        return pre_y

    def _predict_by_group(self, X):
        from copy import deepcopy
        cpset_new = deepcopy(self.loop.cpset)
        se = cpset_new.replace(X)

        res = se.calculate_score(self.expr)
        score, expr01, pre_y = res
        return pre_y

    def predict(self, X):
        if self.expr_type == "group":
            return self._predict_by_group(X)
        else:
            return self._predict_by_single(X)

    def score(self, X, y, scoring):

        y_pred = self.predict(X)

        scoring = check_scoring(self, scoring=scoring)

        if not isinstance(scoring, (list, tuple)):
            scoring = [scoring, ]
        try:
            sc_all = []
            for si in scoring:
                sc = si(self, y, y_pred)
                sc_all.append(sc)

        except (ValueError, RuntimeWarning):

            sc_all = None

        return sc_all

    def cv_result(self, refit=True):
        if self.loop.cpset.cv != 1:
            self.loop.cpset.refit = refit
            return self.loop.cpset.calculate_cv_score(self.best_one.expr)
        else:
            return None


if __name__ == "__main__":
    # data
    from sklearn.datasets import load_boston

    data = load_boston()
    x = data["data"]
    y = data["target"]
    c = [6, 3, 4]

    sl = SymbolLearning(pop=50, gen=3, cal_dim=True, re_hall=2, add_coef=True, cv=2, random_state=1
                        )
    sl.fit(x, y, c=c, x_group=[[1, 3], [0, 2], [4, 7]])
    score = sl.score(x, y, "r2")
    print(sl.expr)




import numbers
import numpy as np
import sympy
from sympy import ImmutableDenseNDimArray
from sympy.tensor.array.arrayop import Flatten

class NewArray(ImmutableDenseNDimArray):
    def __new__(cls, iterable, shape=None, **kwargs):
        return cls._new(iterable, shape, **kwargs)

    def __add__(self, other):
        from sympy.tensor.array.arrayop import Flatten

        if not isinstance(other, NewArray):
            result_list = [i + other for i in Flatten(self)]
            return type(self)(result_list, self.shape)

        if self.shape != other.shape:
            raise ValueError("array shape mismatch")
        result_list = [i + j for i, j in zip(Flatten(self), Flatten(other))]

        return type(self)(result_list, self.shape)

    def __sub__(self, other):

        if not isinstance(other, NewArray):
            result_list = [i - other for i in Flatten(self)]
            return type(self)(result_list, self.shape)

        if self.shape != other.shape:
            raise ValueError("array shape mismatch")
        result_list = [i - j for i, j in zip(Flatten(self), Flatten(other))]

        return type(self)(result_list, self.shape)

    def __rsub__(self, other):

        if not isinstance(other, NewArray):
            result_list = [other - i for i in Flatten(self)]
            return type(self)(result_list, self.shape)
        else:
            raise NotImplementedError("array shape mismatch")

    def __mul__(self, other):

        if isinstance(other, (np.ndarray,NewArray)):
            if self.shape != other.shape:
                raise ValueError("array shape mismatch")
            else:
                result_list = [i * j for i, j in zip(Flatten(self), Flatten(other))]
                return type(self)(result_list, self.shape)
        else:
            result_list = [i * other for i in Flatten(self)]
            return type(self)(result_list, self.shape)

    def __div__(self, other):

        if isinstance(other, (np.ndarray,NewArray)):
            if self.shape != other.shape:
                raise ValueError("array shape mismatch")
            else:
                result_list = [i / j for i, j in zip(Flatten(self), Flatten(other))]
                return type(self)(result_list, self.shape)
        else:
            result_list = [i / other for i in Flatten(self)]
            return type(self)(result_list, self.shape)

    def __rdiv__(self, other):

        if isinstance(other, (np.ndarray,NewArray)):
            if self.shape != other.shape:
                raise ValueError("array shape mismatch")
            else:
                result_list = [i / j for i, j in zip(Flatten(self), Flatten(other))]
                return type(self)(result_list, self.shape)
        else:
            result_list = [other/i for i in Flatten(self)]
            return type(self)(result_list, self.shape)

    def __pow__(self, other):
        return self._eval_power(other)

    def _eval_power(self, other):
        if isinstance(other, (numbers.Real, sympy.Rational, sympy.Float)):
            result_list = [i ** other for i in Flatten(self)]
            return type(self)(result_list, self.shape)
        else:
            raise ValueError("array shape mismatch")

    def __abs__(self):
        result_list = [sympy.Abs(i) for i in Flatten(self)]
        return type(self)(result_list, self.shape)

    def __rpow__(self, other):
        raise ValueError("array shape mismatch")

    def __pos__(self):
        return self

    __radd__ = __add__
    __truediv__ = __div__  # ?
    __rtruediv__ = __rdiv__  # ?

    def as_coefficient(self,_):
        return None


def gsym_map():
    """user's sympy.expr to np.ndarray function"""

    def Flat(x):
        if isinstance(x, (np.ndarray,NewArray)):
            return sum(x)
        else:
            return x

    def Comp(x):
        if isinstance(x, (np.ndarray,NewArray)):
            return np.prod(x)
        else:
            return x

    def Diff(x):
        if isinstance(x, (np.ndarray, NewArray)):
            if x.shape[0] == 2:
                return x[0] - x[1]
            else:
                return x
        else:
            return x

    def Quot(x):
        if isinstance(x, (np.ndarray, NewArray)):
            if x.shape[0] == 2:
                return x[0] / x[1]
            else:
                return x
        else:
            return x

    def Conv(x):
        if isinstance(x, (np.ndarray,NewArray)):
            if x.shape[0] == 2:
                return NewArray((x[1], x[0]),x.shape)
            else:
                return x
        else:
            return x

    def my_abs(x):
        if isinstance(x, (np.ndarray,NewArray)):
            result_list = [sympy.Abs(i) for i in Flatten(x)]
            return NewArray(result_list, x.shape)
        else:
            return sympy.Abs(x)

    def my_sqrt(x):
        return x.__pow__(0.5)

    def my_exp(x):

        if isinstance(x, (np.ndarray,NewArray)):
            result_list = [sympy.exp(i) for i in Flatten(x)]
            return NewArray(result_list, x.shape)
        else:
            return sympy.exp(x)

    def my_ln(x):

        if isinstance(x, (np.ndarray,NewArray)):
            result_list = [sympy.ln(i) for i in Flatten(x)]
            return NewArray(result_list, x.shape)
        else:
            return sympy.ln(x)
    my_log = my_ln

    def my_sin(x):

        if isinstance(x, (np.ndarray,NewArray)):
            result_list = [sympy.sin(i) for i in Flatten(x)]
            return NewArray(result_list, x.shape)
        else:
            return sympy.sin(x)

    def my_cos(x):

        if isinstance(x, (np.ndarray,NewArray)):
            result_list = [sympy.cos(i) for i in Flatten(x)]
            return NewArray(result_list, x.shape)
        else:
            return sympy.cos(x)

    return {"MAdd": Flat, "MMul": Comp, "MSub": Diff, "MDiv": Quot, "Conv": Conv,
            "Self": lambda x_: x_,
            "Abs": my_abs, "exp": my_exp, "ln": my_ln, 'cos': my_cos, 'sin': my_sin, "log":my_log,
            'sqrt': my_sqrt,
            }


# -*- coding: utf-8 -*-

# @Time   : 2019/5/23 20:40
# @Author : Administrator
# @Project : feature_preparation
# @FileName: backforward.py
# @Software: PyCharm

""""""

import numpy as np
from pymatgen.analysis.local_env import VoronoiNN


def _count_i(vo_data, site_ele, mess='face_dist'):
    nei_all_ele = []
    for elei, ele_name in zip(vo_data, site_ele):
        nei_all = []
        for nei in elei:
            nei_name = nei['site'].specie.name
            data_ = nei['poly_info'][mess]
            nei_all.append([nei_name, data_])
        nei_all = np.array(nei_all)
        nei_all_ele.append((ele_name, nei_all))
    return nei_all_ele


def _count_voronoinn(s1, mess='face_dist'):
    """

    :type s1: pymatgen structre type
    """
    vo = VoronoiNN(tol=0.01,
                   allow_pathological=False, weight='solid_angle',
                   extra_nn_info=True, compute_adj_neighbors=True)
    vo_data = vo.get_all_nn_info(s1)
    site_ele = [i.specie.name for i in s1.sites]
    nei_all_ele = _count_i(vo_data, site_ele, mess=mess)
    a_dict = {}
    for i in s1.types_of_specie:
        a_list = []
        for j in nei_all_ele:
            if j[0] == i.name:
                a_list.extend(j[1])
        a_list = np.array(a_list)
        a_dict[i] = a_list
    result_0 = a_dict[s1.types_of_specie[0]]
    number = []
    for ele in s1.types_of_specie:
        b_index = np.where(result_0[:, 0] == ele.name)[0]
        int_ = result_0[b_index, 1:].astype(float)
        number.append(np.sum(int_) / s1.num_sites)
    return number


def count_voronoinn(structures, mess='area'):
    result = [_count_voronoinn(s1, mess=mess) for i, s1 in enumerate(structures)]
    data = np.array(result)
    return data



# -*- coding: utf-8 -*-

# @Time    : 2019/11/12 15:35
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause
# -*- coding: utf-8 -*-

import numbers

import numpy as np
import numpy.core.numeric as numeric
import sympy


def dim_func():
    """dim function """

    def my_abs(dim):

        return dim

    def my_sqrt(dim):

        return dim.__pow__(0.5)

    def my_exp(dim):

        if isinstance(dim, Dim):
            if dim == dless:
                return dless
            else:
                return dnan

    my_log = my_cos = my_sin = my_exp

    my_funcs = {"Abs": my_abs, "exp": my_exp, "log": my_log, 'cos': my_cos, 'sin': my_sin, 'sqrt': my_sqrt}
    return my_funcs


class Dim(numeric.ndarray):
    """re define the Dimension of sympy """

    def __new__(cls, data, dtype=np.float16, copy=True):

        assert isinstance(data, (numeric.ndarray, list))

        arr = numeric.array(data, dtype=dtype, copy=copy)
        try:
            arr.reshape((7, 1))
        except BaseException:
            raise IndexError("the size of datamnist is not 7")

        shape = arr.shape

        ret = numeric.ndarray.__new__(cls, shape, arr.dtype,
                                      buffer=arr,
                                      order='F')
        return ret

    def __init__(self, *_):
        self.unit = ["kg", "m", "s", "A", "K", "mol", "cd"]

    def __add__(self, other):

        if isinstance(other, Dim) and self != other:
            if other == dless:
                return self
            elif self == dless:
                return other
            else:
                return dnan
        elif isinstance(other, Dim) and self == other:
            return self

        elif isinstance(other, (numbers.Real, sympy.Rational, sympy.Float)):
            return self
        else:
            return dnan

    def __sub__(self, other):
        return self + other

    def __pow__(self, other):
        return self._eval_power(other)

    def _eval_power(self, other):
        if isinstance(other, (numbers.Real, sympy.Rational, sympy.Float)):
            return Dim(np.array(self) * other)
        else:
            return dnan

    def __mul__(self, other):
        if isinstance(other, Dim):
            return Dim(np.array(self) + np.array(other))
        elif isinstance(other, (numbers.Real, sympy.Rational, sympy.Float)):
            return self
        else:
            return dnan

    def __div__(self, other):

        if isinstance(other, Dim):
            return Dim(np.array(self) - np.array(other))
        elif isinstance(other, (numbers.Real, sympy.Rational, sympy.Float)):
            return self
        else:
            return dnan

    def __rdiv__(self, other):
        # return other*spath._eval_power(-1)
        if isinstance(other, (numbers.Real, sympy.Rational, sympy.Float)):
            return self.__pow__(-1)
        else:
            return dnan

    def __abs__(self):
        return self

    def __rpow__(self, other):
        return dnan

    def __eq__(self, other):
        return all(np.equal(self, other))

    def __ne__(self, other):
        return not all(np.equal(self, other))

    def __neg__(self):
        return self

    def __pos__(self):
        return self

    @property
    def allisnan(self):
        return all(np.isnan(self))

    @property
    def anyisnan(self):
        return any(np.isnan(self))

    @property
    def isfloat(self):
        return any(np.modf(self)[0])

    @property
    def isinteger(self):
        return not any(np.modf(self)[0])

    def is_same_base(self, others):
        if isinstance(others, Dim):
            npself = np.array(self)
            npothers = np.array(others)
            x1 = np.linalg.norm(npself)
            x2 = np.linalg.norm(npothers)

            if others ** x1 == self ** x2:
                return True
            else:
                return False
        else:
            return False

    __truediv__ = __div__
    __rtruediv__ = __rdiv__
    __radd__ = __add__
    __rsub__ = __sub__
    __rmul__ = __mul__

    def __str__(self):
        strr = "".join(["{}^{}*".format(i, j) for i, j in zip(self.unit, self)])[:-1]

        return strr


dnan = Dim(np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]))
dless = Dim(np.array([0, 0, 0, 0, 0, 0, 0]))

# if __name__=="__main__":
#
#
# ta = time.time()
#
# x = Dim([1, 2, 3, 4, 5, 6, 7])
# a = Dim(np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, ]))
# b = Dim(np.array([1, 1, 1, 1, 1, 1, 1]))

# c = b + 8
# d = 8 + b
# e = b + a
# f = a + b
# g = b + x
#
# g1 = a + dless
# g2 = dless+a
# g3 = dless + x
# g4 = x+dless
# g5 = dless + 1
# g6 = 1 + dless

# c = b - 8
# d = 8 - b
# e = b - a
# f = a - b
# g = b - x
#
# c = b * 8
# d = 8 * b
# e = b * a
# f = a * b
# g = b * x
#
# c = b / 8
# d = 8 / b
# e = b / a
# f = a / b
# g = b / x
#
# c = b ** 8
# d = 8 ** b
# e = b ** a
# f = a ** b
# g = b ** x

# h = abs(b)
# j = abs(a)
#
# h = -b
# j = -a

# xx = np.copy(a)
# print(a == dnan)
# print(a is a)
# print(a != b)
# print(a == 1)
# #
# k = dim_func()["exp"](a)
# l = dim_func()["exp"](b)
# m = dim_func()["exp"](dless)
#
#
# tb = time.time()
# print(tb - ta)



# -*- coding: utf-8 -*-

# @Time    : 2019/11/12 15:13
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause
import operator
import random
from functools import partial

import numpy as np
from deap.base import Fitness, Toolbox
from deap.gp import staticLimit, cxOnePoint, mutNodeReplacement, genHalfAndHalf
from deap.tools import HallOfFame, MultiStatistics, Statistics, initIterate, initRepeat, selTournament
from sklearn.metrics import explained_variance_score, r2_score

from featurebox.combination import creator
from featurebox.combination.common import eaSimple, calculatePrecision, selKbestDim
from featurebox.combination.deapbase import ExpressionTree, ExpressionSet
from featurebox.combination.dictbase import FixedSet, FixedTree, generate_index, cxOnePoint_index, mutUniForm_index
from featurebox.combination.dim import Dim, dnan
from featurebox.tools.tool import time_this_function, parallelize

print(DeprecationWarning("this module 'featurebox.combination' would deprecated in version 0.85. "
                         "please turn to 'featurebox.symbol.'"))


@time_this_function
def mainPart(x_, y_, pset, max_=5, pop_n=100, random_seed=2, cxpb=0.8, mutpb=0.1, ngen=5,
             tournsize=3, max_value=10, double=False, score=None, cal_dim=True, target_dim=None,
             inter_add=True, iner_add=True, random_add=False, store=True):
    """

    Parameters
    ----------
    target_dim
    max_
    inter_add
    iner_add
    random_add
    cal_dim
    score
    double
    x_
    y_
    pset
    pop_n
    random_seed
    cxpb
    mutpb
    ngen
    tournsize
    max_value

    Returns
    -------

    """

    if score is None:
        score = [r2_score, explained_variance_score]

    if cal_dim:
        assert all([isinstance(i, Dim) for i in pset.dim_list]), "all import dim of pset should be Dim object"

    random.seed(random_seed)
    toolbox = Toolbox()

    if isinstance(pset, ExpressionSet):
        PTrees = ExpressionTree
        Generate = genHalfAndHalf
        mutate = mutNodeReplacement
        mate = cxOnePoint
    elif isinstance(pset, FixedSet):
        PTrees = FixedTree
        Generate = generate_index
        mutate = mutUniForm_index
        mate = partial(cxOnePoint_index, pset=pset)

    else:
        raise NotImplementedError("get wrong pset")
    if double:
        Fitness_ = creator.create("Fitness_", Fitness, weights=(1.0, 1.0))
    else:
        Fitness_ = creator.create("Fitness_", Fitness, weights=(1.0,))

    PTrees_ = creator.create("PTrees_", PTrees, fitness=Fitness_, dim=dnan, withdim=0)
    toolbox.register("generate", Generate, pset=pset, min_=1, max_=max_)
    toolbox.register("individual", initIterate, container=PTrees_, generator=toolbox.generate)
    toolbox.register('population', initRepeat, container=list, func=toolbox.individual)
    # def selection
    toolbox.register("select_gs", selTournament, tournsize=tournsize)
    toolbox.register("select_kbest_target_dim", selKbestDim, dim_type=target_dim, fuzzy=True)
    toolbox.register("select_kbest_dimless", selKbestDim, dim_type="integer")
    toolbox.register("select_kbest", selKbestDim, dim_type='ignore')
    # def mate
    toolbox.register("mate", mate)
    # def mutate
    toolbox.register("mutate", mutate, pset=pset)
    if isinstance(pset, ExpressionSet):
        toolbox.decorate("mate", staticLimit(key=operator.attrgetter("height"), max_value=max_value))
        toolbox.decorate("mutate", staticLimit(key=operator.attrgetter("height"), max_value=max_value))
    # def elaluate
    toolbox.register("evaluate", calculatePrecision, pset=pset, x=x_, y=y_, scoring=score[0], cal_dim=cal_dim,
                     inter_add=inter_add, iner_add=iner_add, random_add=random_add)
    toolbox.register("evaluate2", calculatePrecision, pset=pset, x=x_, y=y_, scoring=score[1], cal_dim=cal_dim,
                     inter_add=inter_add, iner_add=iner_add, random_add=random_add)
    toolbox.register("parallel", parallelize, n_jobs=1, func=toolbox.evaluate, respective=False)
    toolbox.register("parallel2", parallelize, n_jobs=1, func=toolbox.evaluate2, respective=False)

    pop = toolbox.population(n=pop_n)

    haln = 5
    hof = HallOfFame(haln)

    stats1 = Statistics(lambda ind: ind.fitness.values[0] if ind and ind.y_dim in target_dim else 0)
    stats1.register("max", np.max)

    stats2 = Statistics(lambda ind: ind.y_dim in target_dim if ind else 0)
    stats2.register("countable_number", np.sum)
    stats = MultiStatistics(score1=stats1, score2=stats2)

    population, logbook = eaSimple(pop, toolbox, cxpb=cxpb, mutpb=mutpb, ngen=ngen, stats=stats,
                                   halloffame=hof, pset=pset, store=store)
    # if not double:
    #     stats1 = Statistics(lambda ind: ind.fitness.values[0])
    #     stats = MultiStatistics(score1=stats1, )
    #     stats.register("avg", np.mean)
    #     stats.register("max", np.max)
    #     population, logbook = eaSimple(pop, toolbox, cxpb=cxpb, mutpb=mutpb, ngen=ngen, stats=stats,
    #                                halloffame=hof, pset=pset)
    # elif double:
    #     stats1 = Statistics(lambda ind: ind.fitness.values[0])
    #     stats2 = Statistics(lambda ind: ind.fitness.values[1])
    #
    #     stats = MultiStatistics(score1=stats1, score2=stats2)
    #     stats.register("avg", np.mean)
    #     stats.register("max", np.max)
    #     population, logbook = multiEaSimple(pop, toolbox, cxpb=cxpb, mutpb=mutpb, ngen=ngen, stats=stats, alpha=alpha,
    #                                         halloffame=hof, pset=pset)
    # else:
    #     raise TypeError
    return population, hof



#!/usr/bin/python3.7
# -*- coding: utf-8 -*-

# @Time   : 2019/7/29 19:47
# @Author : Administrator
# @Software: PyCharm
# @License: BSD 3-Clause

"""
# Just a copy from xenonpy
"""

from itertools import zip_longest

import pandas as pd
from pymatgen import MPRester
from tqdm import tqdm

from featurebox.tools.exports import Store


def data_fetcher(api_key, mp_ids, elasticity=True):
    """fetch file from pymatgen"""
    print('Will fetch %s inorganic compounds from Materials Project' % len(mp_ids))

    def grouper(iterable, n, fillvalue=None):
        """"
        split requests into fixed number groups
        eg: grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
        Collect data_cluster into fixed-length chunks or blocks"""
        args = [iter(iterable)] * n
        return zip_longest(fillvalue=fillvalue, *args)

    # the following props will be fetched
    mp_props = [
        'band_gap',
        'density',
        'volume',
        'material_id',
        'pretty_formula',
        'elements',
        'efermi',
        'e_above_hull',
        'formation_energy_per_atom',
        'final_energy_per_atom',
        'unit_cell_formula',
        'spacegroup',
        'nelements'
    ]
    if elasticity:
        mp_props.append("elasticity")

    entries = []
    mpid_groups = [g for g in grouper(mp_ids, 20)]

    with MPRester(api_key) as mpr:
        for group in tqdm(mpid_groups):
            mpid_list = [ids for ids in filter(None, group)]
            chunk = mpr.query({"material_id": {"$in": mpid_list}}, mp_props)
            try:
                if elasticity:
                    [entry_i.update(entry_i['elasticity']) for entry_i in chunk if 'elasticity' in entry_i]
            except TypeError:
                pass
            entries.extend(chunk)

    df = pd.DataFrame(entries, index=[e['material_id'] for e in entries])
    # df = df.drop('material_id', axis=1)
    df = df.rename(columns={'unit_cell_formula': 'composition'})
    # df = df['volume_per'] = df['volume']/df['nelements']
    df = df.reindex(columns=sorted(df.columns))
    df = df.T

    return df


# def get_ids(api_key="Di2IZMunaeR8vr9w", name_list=None):
#     """
#     support_proprerity = ['energy', 'energy_per_atom', 'volume', 'formation_energy_per_atom', 'nsites',
#     'unit_cell_formula','pretty_formula', 'is_hubbard', 'elements', 'nelements', 'e_above_hull', 'hubbards',
#     'is_compatible', 'spacegroup', 'task_ids',  'band_gap', 'density', 'icsd_id', 'icsd_ids', 'cif',
#     'total_magnetization','material_id', 'oxide_type', 'tags', 'elasticity']
#     """
#     """
#     $gt	>,  $gte >=,  $lt <,  $lte <=,  $ne !=,  $in,  $nin (not in),  $or,  $and,  $not,  $nor ,  $all
#     """
#     m = MPRester(api_key)
#     ids = m.query(criteria={
#         'pretty_formula': {"$in": name_list},
#         # 'nelements': 2,
#         'spacegroup.number': {"$in": [225]},
#         # 'nsites': {"$lt": 5},
#         # 'formation_energy_per_atom': {"$lt": 0},
#         # "elements": {"$in": ["Al", "Co", "Cr", "Cu", "Fe", 'Ni'], "$all": "O"},
#         # "elements": {"$in": list(combinations(["Al", "Co", "Cr", "Cu", "Fe", 'Ni'], 5))}
#     }, properties=["material_id"])
#     print("number %s" % len(ids))
#     return ids

def get_ids(api_key="Di2IZMunaeR8vr9w", name_list=None):
    """
    support_proprerity = ['energy', 'energy_per_atom', 'volume', 'formation_energy_per_atom', 'nsites',
    'unit_cell_formula','pretty_formula', 'is_hubbard', 'elements', 'nelements', 'e_above_hull', 'hubbards',
    'is_compatible', 'spacegroup', 'task_ids',  'band_gap', 'density', 'icsd_id', 'icsd_ids', 'cif',
    'total_magnetization','material_id', 'oxide_type', 'tags', 'elasticity']
    """
    """
    $gt	>,  $gte >=,  $lt <,  $lte <=,  $ne !=,  $in,  $nin (not in),  $or,  $and,  $not,  $nor ,  $all	
    """
    m = MPRester(api_key)
    ids = m.query(criteria={
        # 'pretty_formula': {"$in": name_list},
        'nelements': {"$lt": 5, "$gt": 3},
        # 'spacegroup.number': {"$in": [225]},
        'nsites': {"$lt": 20},
        'formation_energy_per_atom': {"$lt": 0},
        # "elements": {"$in": ["Al", "Co", "Cr", "Cu", "Fe", 'Ni'], "$all": "O"},
        # "elements": {"$in": list(combinations(["Al", "Co", "Cr", "Cu", "Fe", 'Ni'], 5))}
    }, properties=["material_id"])
    print("number %s" % len(ids))
    return ids


if __name__ == "__main__":
    list1 = list(
        ['CsCl', 'CsBr', 'CsI', 'CsSb', 'LiF', 'KF', 'RbF', 'CsF', 'MgO', 'CdO', 'MnO', 'VO', 'CaO', 'SrO', 'BaO',
         'EuO', 'ScN', 'YN', 'ErN', 'HoN', 'DyN', 'GdN', 'EuN', 'CeN', 'LiCl', 'TlCl', 'AgCl', 'NaCl', 'RbCl', 'LiBr',
         'TlBr', 'AgBr', 'NaBr', 'KBr', 'RbBr', 'MgSe', 'PbSe', 'CaSe', 'SrSe', 'BaSe', 'YbSe', 'EuSe', 'SmSe', 'PbS',
         'MnS', 'CaS', 'SrS', 'BaS', 'YbS', 'EuS', 'SmS', 'LiI', 'TlI', 'NaI', 'KI', 'RbI', 'YbAs', 'TmAs', 'DyAs',
         'GdAs', 'NdAs', 'SmAs', 'PrAs', 'SmP', 'AsTe', 'GeTe', 'SnTe', 'PbTe', 'CaTe', 'SrTe', 'BaTe', 'YbTe', 'ErTe',
         'GdTe', 'EuTe', 'SmTe', 'LaSb', 'YbSb', 'SmSb', 'PrSb', 'NaF', 'KCl', 'CuBr', 'BeSe', 'ZnSe', 'CdSe', 'HgSe',
         'BeS', 'ZnS', 'CdS', 'AlAs', 'AlP', 'BeTe', 'ZnTe', 'CdTe', 'HgTe', 'AlSb', 'BN', 'SiC3c', 'GaAs', 'InAs',
         'BP', 'GaP', 'InP', 'GaSb', 'InSb', 'CuCl', 'HgS', 'CuI', 'MnTe', 'AgI', 'ZnS', 'ZnSe', 'ZnO', 'AlN', 'GaN',
         'MgTe', 'BeO', 'BN', 'InN', 'SiC', 'MnS'])
    idss = get_ids(api_key="Di2IZMunaeR8vr9w", name_list=list1)
    idss1 = [i['material_id'] for i in idss]
    dff = data_fetcher("Di2IZMunaeR8vr9w", idss1, elasticity=False)
    st = Store(r"C:\Users\Administrator\Desktop")
    st.to_csv(dff, "id_structure")



import numpy as np
import sympy
from sympy import Function

from featurebox.symbol.functions.dimfunc import Dim
from featurebox.symbol.functions.gsymfunc import NewArray


def newfuncV(operation, arity=1, name="Fc"):
    """

    Parameters
    ----------
    operation:callable
        the detail of opearation only accept +,-,*,/,abs,-(negative),x^n
    arity: int
        the arity of opeaation
    name:str
        name
    Returns
    -------
    dict
    """
    func = Function(name)

    def npf(*x):
        return operation(*x)

    def gsymf(*x):
        return operation(*x)

    def dimf(*dim):
        return operation(*dim)

    return {"func": func, "arity": arity, "name": name, "np_func": npf,
            "dim_func": dimf, "sym_func": gsymf}


def newfuncD(operation, name="Fc", keep=True, is_jump=False, check=True):
    """

    Parameters
    ----------
    operation : callable
        the detail of opearation only accept +,-,*,/,-(negative),x^n
    name:str
        name
    keep:bool
        the group size after this function. true is the input size,and false is 1.
    is_jump:bool
        the bool means the rem and rem_dim can be treat 2+ domension problems or not.
    check:bool
        check the function building
    """

    func = Function(name)
    func.is_jump = is_jump
    func.keep = keep

    def npf(x):
        if isinstance(x, np.ndarray):
            if x.ndim == 2:
                if x.shape[0] == 2:
                    res = operation(x[0], x[1])
                    if keep:
                        assert isinstance(res, tuple)
                    return np.array(res)
                else:  # >=3
                    if is_jump:
                        return x
                    else:
                        return np.array(operation(*x))
            else:  # 1
                return x
        else:  # number
            return x

    def gsymf(x):
        if isinstance(x, (np.ndarray, NewArray)):
            if x.shape[0] == 2:
                res = operation(x[0], x[1])
                if keep:
                    return NewArray(res)
                else:
                    return res
            else:  # >=3
                if is_jump:
                    return x
                else:
                    res = operation(*x)
                    if keep:
                        return NewArray(res)
                    else:
                        return res
        else:  # 1
            return x

    def dimf(dim):

        if isinstance(dim, Dim):
            if dim.ndim == 1:
                return dim
            elif dim.shape[0] == 2:
                if keep:
                    return Dim(operation(*dim))
                else:
                    return Dim(operation(*dim))
            else:
                if is_jump:
                    return dim
                else:
                    return Dim(operation(*dim))
        else:  # number
            return dim

    res = {"func": func, "arity": 1, "name": name, "np_func": npf, "dim_func": dimf, "sym_func": gsymf}

    if check:
        check_funcD(res)

    return res


def check_funcD(funcs, self_grpup=2):
    """self_group>=2"""

    def test_npf():
        a = 1
        b = np.array([1, 2, 3, 4, 5])
        c = np.array([[1, 2, 3, 4, 5], [1, 2, 3, 4, 5]])
        d = np.array([[1, 2, 3, 4, 5], [1, 2, 3, 4, 5], [1, 2, 3, 4, 5]])
        e = np.array([[1, 2, 3, 4, 5] * self_grpup])
        np_func = funcs["np_func"]
        np_func(a)
        np_func(b)
        np_func(c)
        s = np_func(d)
        s = np_func(e)

    def test_dim_func():
        a = 1
        b = Dim(np.array([1, 2, 3, 4, 5]))
        c = Dim(np.array([[1, 2, 3, 4, 5], [1, 2, 3, 4, 5]]))
        d = Dim(np.array([[1, 2, 3, 4, 5], [1, 2, 3, 4, 5], [1, 2, 3, 4, 5]]))
        e = Dim(np.array([[1, 2, 3, 4, 5] * self_grpup]))
        dim_func = funcs["dim_func"]
        dim_func(a)
        dim_func(b)
        dim_func(c)
        s = dim_func(d)
        s = dim_func(e)

    def test_sym_func():
        a = 1
        b = sympy.Symbol("x")
        c = NewArray((sympy.Symbol("x"), sympy.Symbol("y")))
        d = NewArray((sympy.Symbol("x"), sympy.Symbol("x"), sympy.Symbol("z")))
        e = NewArray([sympy.Symbol("x")] * self_grpup)
        sym_func = funcs["sym_func"]
        sym_func(a)
        sym_func(b)
        sym_func(c)
        s = sym_func(d)
        s = sym_func(e)

    test_npf()
    test_dim_func()
    test_sym_func()


if __name__ == "__main__":
    def funcs(*arg):
        return sum(arg)


    newfuncD(funcs, name="Fc", keep=False, is_jump=False, check=True)


    def funcs(*arg):
        return arg[0] ** 2 + arg[1]


    newfuncD(funcs, name="Fc", keep=False, is_jump=True, check=True)


    def funcs(*arg):
        return tuple([2.5 * argi ** 2 for argi in arg])


    newfuncD(funcs, name="Fc", keep=True, is_jump=False, check=True)



# -*- coding: utf-8 -*-

# @Time   : 2019/7/13 19:27
# @Author : Administrator
# @Project : feature_toolbox
# @FileName: mmgs.py
# @Software: PyCharm

"""
UGS
this is a union select method for feature subsets.
key:
1.gather performance of different model
2.raise the best one from similar learning performance subsets
3.rank the raised subsets and penalty the size of subsets

node == feature subset
"""
import copy
import functools
import itertools
import numbers
import warnings
from collections import Counter
from copy import deepcopy
from functools import partial
from operator import itemgetter

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from sklearn import metrics, preprocessing
from sklearn.cluster import DBSCAN
from sklearn.metrics import get_scorer
from sklearn.model_selection import KFold, GridSearchCV, StratifiedKFold
from sklearn.utils import check_X_y, check_random_state
from sklearn.utils.multiclass import type_of_target

from featurebox.tools.tool import parallelize

warnings.filterwarnings("ignore")


def displacement(binary_distance, print_noise=0.001):
    rd = check_random_state(0)
    q = rd.random_sample(binary_distance.shape) * print_noise / 10
    binary_distance = binary_distance + q

    indexs = np.argwhere(binary_distance <= 0)
    indexs = indexs[np.where(indexs[:, 0] > indexs[:, 1])]
    t = rd.random_sample(indexs.shape[0]) * print_noise / 20
    binary_distance[indexs[:, 0], indexs[:, 1]] = t
    binary_distance[indexs[:, 1], indexs[:, 0]] = t
    return binary_distance


def cluster_printing(slices, node_color, edge_color_pen=0.7, binary_distance=None, print_noise=0.001,
                     node_name=None, highlight=None):
    """

        Parameters
        ----------
        highlight:list
            change shape
        slices: list
            the lists of the index of feature subsets, each feature subset is a node.
            Examples 3 nodes
            [[1,4,5],[1,4,6],[1,2,7]]
        node_color: np.ndarray 1D, list, the same size as slices
            the label to classify the node
        edge_color_pen: int
            the transparency of edge between node
        binary_distance: np.ndarray
            distance matrix for each pair node
        print_noise: int
            add noise for less printing overlap
        node_name: list
            x_name of node
    """
    plt.figure()
    g = nx.Graph()

    def _my_ravel(data_cof):
        for i in range(data_cof.shape[0]):
            for k in range(i + 1, data_cof.shape[0]):
                yield i, k, data_cof[i, k]

    distances = displacement(binary_distance, print_noise=print_noise)

    distance_weight = list(_my_ravel(distances))
    g.add_weighted_edges_from(distance_weight)
    # edges=nx.get_edge_attributes(g, 'weight').items()
    edges, weights = zip(*nx.get_edge_attributes(g, 'weight').items())
    weights = weights / max(weights)
    pos = nx.layout.kamada_kawai_layout(g)  # calculate site

    if node_name is None:
        le = binary_distance.shape[0] or len(slices)
        lab = {i: i for i in range(le)}
    else:
        assert binary_distance.shape[0] or len(slices) == len(node_name)
        if isinstance(node_name, list) and isinstance(node_name[0], list):
            strr = ","
            node_name = [strr.join(i) for i in node_name]
        lab = {i: j for i, j in enumerate(node_name)}

    nodesize = [600] * distances.shape[0]
    node_edge_color = ["w"] * distances.shape[0]
    if highlight:
        for i in highlight:
            node_edge_color[i] = "aqua"
        for i in highlight:
            nodesize[i] *= 1.3

    mp = {-1: 'gray', 0: "mediumpurple", 1: 'seagreen', 2: 'goldenrod', 3: 'deeppink', 4: "chocolate",
          5: "lightseagreen", }
    node_color = list(map(lambda x: mp[x], node_color))

    nx.draw(g, pos, edgelist=edges, edge_color=np.around(weights, decimals=3) ** edge_color_pen,
            edge_cmap=plt.cm.Blues_r, edge_labels=nx.get_edge_attributes(g, 'weight'), edge_vmax=0.7, width=weights,
            labels=lab, font_size=12,
            node_color=np.array(node_color), vmin=-1, max=10,
            edgecolors=node_edge_color, linewidths=1,
            node_cmap=plt.cm.Paired, node_size=nodesize,
            )

    plt.show()


def score_group(cl_data, label0):
    """
    score the group results

    Parameters
    ----------
    cl_data: np.ndarray
        cluster import data_cluster
    label0: np.ndarray
        node distance matrix

    Returns
    -------
    group_score_i: list
        score for this group_i result
    """
    # sdbw = SDbw(cl_data, label0)
    # res = sdbw.result()
    if max(list(Counter(label0).items()), key=itemgetter(1))[1] > 0.8 * len(label0):
        res = -np.inf
    elif 2 <= max(label0) + 1 < len(label0):
        # res = -metrics.davies_bouldin_score(cl_data, label0)
        res = -metrics.calinski_harabasz_score(cl_data, label0)

    else:
        res = -np.inf
    return res


def sc(epsi, distances):
    """
    cluster the node and get group results

    Parameters
    ----------
    epsi: int
        args for DBSCAN
    distances: np.ndarray
        distance matrix

    Returns
    -------
    group_i: list
        Examples 4 groups for 8 node
        [[0,4,5],[2,3,6],[7],[1]]
    label_i: np.ndarray, 1D
        label for node, the same label is the same group
        Examples 4 groups for 8 node
        [0,3,1,1,0,0,1,2]
    label_0: np.ndarray, 1D
        label for node, single groups are separated
        [0,3,1,1,0,0,1,2]
    """
    db = DBSCAN(algorithm='auto', eps=epsi, metric='precomputed',
                metric_params=None, min_samples=2, n_jobs=None, p=None)
    distances /= np.max(distances)
    db.fit(distances)

    label_i = db.labels_
    set_label = list(set(label_i))
    group_i = [[i for i in range(len(label_i)) if label_i[i] == aim] for aim in set_label]

    label_0 = deepcopy(label_i)
    for j in range(len(label_0)):
        if label_0[j] == -1:
            label_0[j] = max(label_0) + 1

    return group_i, label_i, label_0


class GS(object):
    """
    grouping selection

    calculate the predict_y of base estimator on node
    calculate the distance of predict_y
    cluster the nodes by distance and get groups
    select the candidate nodes in each groups with penalty of size of node (feature number)
    rank the candidate nodes


    """

    def __init__(self, estimator, slices, estimator_i=0, n_jobs=2, batch_size=1):
        """

        Parameters
        ----------
        estimator : list
            list of base estimator or GridSearchCV from sklearn
        slices: list
            the lists of the index of feature subsets, each feature subset is a node,each int is the index of X
            Examples 3 nodes
            [[1,4,5],[1,4,6],[1,2,7]]
        estimator_i: int
            default index of estimator
        """
        self.slices = slices
        self.estimator_i = estimator_i
        if isinstance(estimator, list):
            self.estimator = estimator
        else:
            self.estimator = [estimator, ]
        scorer_all = []
        cv_all = []
        for i in self.estimator:
            assert isinstance(i, GridSearchCV)
            scorer_all.append(get_scorer(i.scoring))
            cv_all.append(get_scorer(i.cv))
        self.predict_y = []  # changed with estimator_i
        self.n_jobs = n_jobs
        self.batch_size = batch_size
        scorer = scorer_all[0]
        scorer_func = scorer._score_func
        self.metrics_method = scorer_func
        self.cv = copy.deepcopy(cv_all[0])

    def fit(self, x, y):
        """
        Parameters
        ----------
        x: np.ndarray
        y: np.ndarray, 1D
        """
        x, y = check_X_y(x, y, "csc")
        self.x0 = x
        self.y0 = y
        cv = self.cv
        if isinstance(cv, numbers.Integral):
            if (type_of_target(y) in ('binary', 'multiclass')):
                cv = StratifiedKFold(cv)
            else:
                cv = KFold(cv)
        self.kf = list(cv.split(x))

    @functools.lru_cache(1024)
    def _cv_predict(self, slices_i, estimator_i):
        """ Fit y with in slices_i,resample for n_resample times """
        estimator = self.estimator[estimator_i]
        assert isinstance(estimator, GridSearchCV)
        x0 = self.x0
        y0 = self.y0
        kf = self.kf
        slices_i = list(slices_i)
        if len(slices_i) <= 1:
            raise ValueError("feature number should large than 1")
        else:
            data_x0 = x0[:, slices_i]
            estimator.fit(data_x0, y0)
            x, y = data_x0, y0

            s_estimator = copy.deepcopy(estimator.best_estimator_)

            y_test_predict_all = []
            for train, test in kf:
                X_train, X_test, y_train, y_test = x[train], x[test], y[train], y[test]
                s_estimator.fit(X_train, y_train)
                y_test_predict = s_estimator.predict(X_test)
                y_test_predict_all.append(y_test_predict)

        return y_test_predict_all

    def _cv_predict_all(self, slices=None, estimator_i=0):
        """ calculate binary distance of 2 nodes """

        self.estimator_i = estimator_i if isinstance(estimator_i, int) else self.estimator_i
        n_jobs = self.n_jobs
        batch_size = self.batch_size
        slices = slices if slices else self.slices

        ret = self.check_prop("cv_predict_all", estimator_i=self.estimator_i, slices=slices)

        if ret is not None:
            pass
        else:
            cal_score = partial(self.predict)
            ret = parallelize(n_jobs=n_jobs, func=cal_score, iterable=slices, batch_size=batch_size)

            self.add_prop("cv_predict_all", estimator_i=self.estimator_i, slices=slices, values=ret)

        self.slices = slices

        return ret

    def predict(self, slice_i):
        """change type """
        estimator_i = self.estimator_i
        return self._cv_predict(tuple(slice_i), estimator_i)

    def cv_score(self, slices_i):

        y_test_predict_all = self.predict(slices_i)
        test_index = [i[1] for i in self.kf]
        y_test_true_all = [self.y0[_] for _ in test_index]
        score = [self.metrics_method(i, j) for i, j in zip(y_test_true_all, y_test_predict_all, )]
        score_mean = np.mean(score)

        return score_mean

    def cv_score_all(self, slices=None, estimator_i=0):
        """score all node with r2

        Parameters
        ----------
        slices : list, or None, default spath.slices
            change to new slices to calculate
            the lists of the index of feature subsets, each feature subset is a node,each int is the index of X
            Examples 3 nodes
            [[1,4,5],[1,4,6],[1,2,7]]
        estimator_i: int, default spath.estimator_i
            change to the estimator_i to calculate

        Returns
        ----------
            score_mean_std: nd.ndarray 2D
            the mean and std

        """

        self.estimator_i = estimator_i if isinstance(estimator_i, int) else self.estimator_i
        n_jobs = self.n_jobs
        batch_size = self.batch_size
        slices = slices if slices else self.slices

        ret = self.check_prop("cv_score_all", estimator_i=self.estimator_i, slices=slices)

        if ret is not None:
            pass
        else:
            cal_score = partial(self.cv_score)
            ret = parallelize(n_jobs=n_jobs, func=cal_score, iterable=slices, batch_size=batch_size)

            self.add_prop("cv_score_all", estimator_i=self.estimator_i, slices=slices, values=ret)

        self.slices = slices
        return np.array(ret)

    def cal_y_distance(self, slice1):
        """ calculate binary distance of 2 nodes """
        set1 = set(slice1)

        test_index = [i[1] for i in self.kf]
        y_true_all = [self.y0[_] for _ in test_index]

        y_1_all = self.predict(set1)
        y_2_all = y_1_all

        # distance = [spatial.distance.euclidean(i, j) for i, j, k in zip(y_true_all, y_1_all, y_2_all)]
        distance = [1 - self.metrics_method(i, j) for i, j, k in zip(y_true_all, y_1_all, y_2_all)]
        distance = np.mean(distance)
        distance = distance if distance >= 0 else 0
        return distance

    def cal_y_distance_all(self, slices=None, estimator_i=0):
        """ calculate binary distance of 2 nodes """

        self.estimator_i = estimator_i if isinstance(estimator_i, int) else self.estimator_i
        n_jobs = self.n_jobs
        batch_size = self.batch_size
        slices = slices if slices else self.slices

        ret = self.check_prop("cal_y_distance_all", estimator_i=self.estimator_i, slices=slices)

        if ret is not None:
            pass
        else:
            cal_score = partial(self.cal_y_distance)
            ret = parallelize(n_jobs=n_jobs, func=cal_score, iterable=slices, batch_size=batch_size)

            self.add_prop("cal_y_distance_all", estimator_i=self.estimator_i, slices=slices, values=ret)

        self.slices = slices
        return np.array(ret)

    def cal_binary_distance(self, slice1, slice2):
        """ calculate binary distance of 2 nodes """
        set1 = set(slice1)
        set2 = set(slice2)

        test_index = [i[1] for i in self.kf]
        y_true_all = [self.y0[_] for _ in test_index]
        y_1_all = self.predict(set1)
        y_2_all = self.predict(set2)

        # distance = [spatial.distance.euclidean(j, k) for i, j, k in zip(y_true_all, y_1_all, y_2_all)]
        distance = [1 - self.metrics_method(j, k) for i, j, k in zip(y_true_all, y_1_all, y_2_all)]
        distance = np.mean(distance)
        distance = distance if distance >= 0 else 0
        return distance

    def cal_binary_distance_all(self, slices=None, estimator_i=0):
        """ calculate the distance matrix of slices """
        self.estimator_i = estimator_i if isinstance(estimator_i, int) else self.estimator_i
        n_jobs = self.n_jobs
        batch_size = self.batch_size
        slices = slices if slices else self.slices

        ret = self.check_prop("cal_binary_distance_all", estimator_i=self.estimator_i, slices=slices)

        if ret is not None:
            pass
        else:
            cal_binary_distance = partial(self.cal_binary_distance)
            slices_cuple = list(itertools.product(slices, repeat=2))
            ret = parallelize(n_jobs=n_jobs, func=cal_binary_distance, iterable=slices_cuple, respective=True,
                              batch_size=batch_size)
            ret = np.reshape(ret, (len(slices), len(slices)), order='F')

            self.add_prop("cal_binary_distance_all", estimator_i=self.estimator_i, slices=slices, values=ret)

        self.slices = slices
        return ret

    def check_prop(self, prop, estimator_i=0, slices=None):
        if slices == self.slices:
            if hasattr(self, "".join(["result", prop, "_", "%s" % estimator_i])):
                return getattr(self, "".join(["result", prop, "_", "%s" % estimator_i]), None)

    def add_prop(self, prop, slices, estimator_i, values):
        if hasattr(self, "".join(["result", prop, "_", "%s" % estimator_i])) and slices == self.slices:
            pass
        else:
            setattr(self, "".join(["result", prop, "_", "%s" % estimator_i]), values)

    # @time_this_function
    def cal_group(self, eps=None, printing=False, slices=None, estimator_i=0,
                  print_noise=0.1, node_name=None, pre_binary_distance_all=None):
        """

        Parameters
        ----------

        eps: int
            args for DBSCAN
        printing: bool
            default,True for GS and False for UGS
        slices : list, or None, default spath.slices
            change to new slices to calculate
            the lists of the index of feature subsets, each feature subset is a node,each int is the index of X
            Examples 3 nodes
            [[1,4,5],[1,4,6],[1,2,7]]
        estimator_i: int, default spath.estimator_i
            change to the estimator_i to calculate
        print_noise: int
            add noise for less printing overlap
        node_name: list of str
            x_name of node, be valid for printing is True
        pre_binary_distance_all
            pre-calculate binary_distance_all ,please make sure the slices and estimator_i are same
            as the slices and estimator_i in pre-calculate binary_distance_all

        Returns
        -------
        group: list
            group results, the result of groups is unique
            Examples 4 groups for 8 node
            [[0,4,5],[2,3,6],[7],[1]]
        """
        self.estimator_i = estimator_i if isinstance(estimator_i, int) else self.estimator_i
        self.slices = slices if slices else self.slices
        slices = self.slices

        if isinstance(print_noise, (float, int)) and 0 < print_noise <= 1:
            pass
        else:
            raise TypeError("print_noise should be in (0,1]")

        if not isinstance(pre_binary_distance_all, np.ndarray):
            binary_distance = self.cal_binary_distance_all(slices, self.estimator_i)
        else:
            binary_distance = pre_binary_distance_all

        pre_y = self._cv_predict_all(slices, self.estimator_i)
        pre_y = np.array([np.concatenate([i.ravel() for i in pre_yi]).T for pre_yi in pre_y])

        distances = binary_distance
        if eps:
            group, label, label_sep = sc(eps, distances)
        else:
            group = None
            group_score = -np.inf
            label = [[1] * len(self.slices)]

            for epsi in np.arange(0.05, 0.95, 0.01):
                group_i, label_i, label_sep_i = sc(epsi, distances)
                group_score_i = score_group(pre_y, label_sep_i)
                print(label_i, group_score_i)
                if group_score_i > group_score:
                    group_score = group_score_i
                    # print(epsi, group_score)
                    group = group_i
                    label = label_i

        if printing:
            self.cluster_print(binary_distance, print_noise=print_noise, node_name=node_name, label=label)
        group_last = group.pop()
        group += [[i] for i in group_last]
        self.group = group
        return group

    def cluster_print(self, binary_distance, label=None, eps=0.1, print_noise=0.001, node_name=None, highlight=None):
        """ print temporary"""
        if label is None:
            group, label, label_sep = sc(eps, binary_distance)
        else:
            label = label
        cluster_printing(slices=self.slices, binary_distance=binary_distance,
                         print_noise=print_noise, node_name=node_name,
                         node_color=label, highlight=highlight)

    # def distance_print(spath):

    def select_gs(self, alpha=0.01):
        """

        Parameters
        ----------
        alpha: int
            penalty coefficient

        Returns
        -------
        score_select: list
            selected node score
        selected: list
            selected node in import
        site_select: list
            selected node number in import
        """

        score_all = np.array(self.cv_score_all(self.slices))

        score = score_all[:, 0]
        std = score_all[:, 1]
        max_std = np.max(std)
        t = 2
        m = np.array([len(i) for i in self.slices])

        score = score * (1 - std / max_std) - alpha * (np.exp(m - t) + 1)
        score = preprocessing.minmax_scale(score)

        score_select, selected, site_select = self._select(self.slices, self.group, score, fliters=False, )
        return score_select, selected, site_select

    @staticmethod
    def _select(slices, group, score, fliters=False):
        """select the maximum, greater is better. Not suit for minimum"""
        score_groups = [[score[i] for i in slicei_group] for slicei_group in group]
        select = [np.argmax(i) for i in score_groups]  # select in group_i
        site_select = [i[_] for _, i in zip(select, group)]  # site
        site_select = list(set(site_select)) if fliters else site_select
        score_select = [score[_] for _ in site_select]  # score_select
        selected = [slices[_] for _ in site_select]  # select

        result = sorted(zip(score_select, selected, site_select), key=itemgetter(0), reverse=True)
        result = list(zip(*result))
        return result


class UGS(GS):
    """
    union grouping selection

    calculate the predict_y  on node, for each base estimator
    calculate the distance of predict_y, for each base estimator
    cluster the nodes by distance and get groups, for each base estimator
    unite groups of base estimators to tournament groups
    select the candidate nodes in each groups with penalty of size of node (feature number)
    rank the candidate nodes


    """

    def __init__(self, estimator, slices, estimator_n=None, n_jobs=2, estimator_i=0, batch_size=1):
        """

        Parameters
        ----------
        estimator : list
            list of base estimator or GridSearchCV from sklearn
        slices: list
            the lists of the index of feature subsets, each feature subset is a node,each int is the index of X
            Examples 3 nodes
            [[1,4,5],[1,4,6],[1,2,7]]
        estimator_n: list
            default indexes of estimator
        """
        super().__init__(estimator, slices, estimator_i, n_jobs=n_jobs, batch_size=batch_size)
        if estimator_n is None:
            self.estimator_n = list(range(len(estimator)))
        else:
            self.estimator_n = estimator_n
        assert len(self.estimator) >= 2
        assert len(self.estimator_n) >= 2

    def cal_t_group(self, eps=None, printing=False, pre_group=None):
        """

        Parameters
        ----------
        eps: int
            args for DBSCAN
        printing: bool
            draw or not, default, False
        pre_group: None or list of different estimator's groups
            the sort of pre_group should match to spath.estimator !
            pre-calculate results by spath.cal-group for all base estimator, to escape double counting


        Returns
        -------
        tn_group: list
            the element of tournament groups can be repeated
            Examples
            [[1,2,3],[3,4,5],[1,6,7],[2,3]]

        """

        slices = [tuple(_) for _ in self.slices]
        if not pre_group:
            group_result = [self.cal_group(estimator_i=i, eps=eps, printing=printing, print_noise=0.01, node_name=None)
                            for i in self.estimator_n]
        else:
            assert len(pre_group) == self.estimator_n, "the size of pre_group should is the number fo estimator!"
            group_result = pre_group
        for group in group_result:
            single = group.pop()
            single = [[_] for _ in single]
            group.extend(single)
        tn_group = []
        for slicei in range(len(slices)):
            slicei_group = set()
            for group in group_result:
                for groupi in group:
                    if slicei in groupi:
                        slicei_group.update(groupi)
            slicei_group = list(slicei_group)
            tn_group.append(slicei_group)
        # todo hebing ?
        self.group = tn_group
        return tn_group

    def select_ugs(self, alpha=0.01):
        """

        Parameters
        ----------
        alpha: int
            penalty coefficient

        Returns
        -------
        score_select: list
            selected node score
        selected: list
            selected node in import
        site_select: list
            selected node number in import

        """

        score_all = np.array([self.cv_score_all(estimator_i=i)[:, 0] for i in self.estimator_n])

        score = np.mean(np.array(score_all), axis=0)
        std = np.std(np.array(score_all), axis=0)
        max_std = np.max(std)
        t = 2
        m = np.array([len(i) for i in self.slices])

        score = score * (1 - std / max_std) - alpha * (np.exp(m - t) + 1)
        score = preprocessing.minmax_scale(score)
        score_select, selected, site_select = self._select(self.slices, self.group, score, fliters=True)

        result = sorted(zip(score_select, selected, site_select), key=itemgetter(0), reverse=True)
        result = list(zip(*result))
        return result



import time

from numpy import random
from sklearn.datasets import load_boston

from featurebox.symbol.base import SymbolSet, SymbolTree, CalculatePrecisionSet
from featurebox.symbol.functions.dimfunc import dless, Dim
from featurebox.symbol.functions.npfunc import np_map
from featurebox.symbol.calculation.scores import compile_context, calculate_y


if __name__ == "__main__":
    # data
    data = load_boston()
    x = data["data"]
    y = data["target"]
    c = [6, 3, 4]
    # unit
    from sympy.physics.units import kg

    x_u = [kg] * 13
    y_u = kg
    c_u = [dless, dless, dless]

    x, x_dim = Dim.convert_x(x, x_u, target_units=None, unit_system="SI")
    y, y_dim = Dim.convert_xi(y, y_u)
    c, c_dim = Dim.convert_x(c, c_u)

    t = time.time()

    # symbolset
    pset0 = SymbolSet()
    pset0.add_features(x, y, x_dim=x_dim, y_dim=y_dim, x_group=[[1, 2], [3, 4, 5], [6, 7]],
                       feature_name=["Ss%i" % i for i in range(13)])
    pset0.add_constants(c, c_dim=c_dim, c_prob=None)
    pset0.add_operations(power_categories=(2, 3, 0.5),
                         categories=("Add", "Mul", "Sub", "Div", "ln"),
                         self_categories=None)

    random.seed(2)
    z = time.time()
    sl = [SymbolTree.genGrow(pset0, 3, 4) for _ in range(500)]
    a = time.time()
    # sli =" MAdd(Sub(Add(Mul(x0, gx1), exp(x10)), Mul(Conv(Add(x0, gx0)), Mul(x6, MAdd(x10)))))"
    # sl =["MAdd(gx1 * x11 * (-x0 + x11) * MAdd(gx1))"]

    # sl = [compile_context(sli, pset0.context, pset0.gro_ter_con) for sli in sl]
    # sl = [simple(sli.args[0], pset0.gro_ter_con) for sli in sl if len(sli.args)>0]
    c = 1
    a = time.time()
    pset0 = CalculatePrecisionSet(pset0, scoring=None, score_pen=(1,), filter_warning=True, cal_dim=False,
                                  dim_type=None,
                                  fuzzy=False, add_coef=False, inter_add=True,
                                  inner_add=False, n_jobs=1, batch_size=20,
                                  tq=True)
    from sys import getsizeof
    import numpy as np
    T100 = getsizeof(sl)
    psize = getsizeof(pset0)
    # print(T100,psize)
    for i in sl:
        b = time.time()
        i0 = compile_context(i, pset0.context, pset0.gro_ter_con)
        r2 = time.time()
        # pprint(i0,pset0)
        calculate_y(i0, pset0.data_x,pset0.y, pset0.terminals_and_constants_repr, add_coef=True,
                    filter_warning=True, inter_add=True, inner_add=True, np_maps=np_map())
        # i = pset0.calculate_detail(i)
        # j = simple(i.coef_expr, pset0.gro_ter_con)
        # print(i)
        # print(repr(i))
        # print(i0)
        # ppprint(i, pset0, feature_name=True)
        if y[0] is None:
            c = c + 1

        r3 = time.time()
    #     print(r2 - b, "com")
    #     print(r3-r2,"cal")
    # e = time.time()
    # print(e-a)



import os

import pandas as pd

pa = os.path.abspath(__file__)
pa = os.path.split(pa)[0]
pa = os.path.join(pa, r"element_table.xlsx")

element_table = pd.read_excel(pa,
                              header=7, skiprows=0, index_col=0)



import numpy as np
import torch
from matplotlib.pyplot import imshow
from torch import nn
from torchvision.models.vgg import make_layers

from featurebox.cam.cam import GradCRM
from featurebox.cam.fig import normalize, trans_to_numpy, visualize
from featurebox.data.datasets import CAMData


class VGG(nn.Module):

    def __init__(self, features, num_classes=1000, init_weights=True):
        super(VGG, self).__init__()
        self.features = features
        self.avgpool = nn.AdaptiveAvgPool2d((3, 3))
        self.classifier = nn.Sequential(
            nn.Linear(256 * 3 * 3, 1000),
            nn.ReLU(True),
            nn.Dropout(),
            nn.Linear(1000, 1000),
            nn.ReLU(True),
            nn.Dropout(),
            nn.Linear(1000, 1),
        )
        if init_weights:
            self._initialize_weights()

    def forward(self, x):
        x = self.features(x)
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)


if __name__ == "__main__":

    sample = 100
    torch.set_default_tensor_type(torch.FloatTensor)
    camdata = CAMData(20, 20, sample, random_state=1, question="reg", inter=False, re_range=(-0.5, 1))
    data = camdata.gen_data(featurize_x=True)
    x = data.x
    y = data.y
    x = x.astype(np.float32)
    y = y.astype(np.float32)
    x = x[np.newaxis, :, :, :]
    x = np.repeat(x, 3, axis=0)

    x = x.transpose((3, 0, 1, 2))

    t_x = torch.from_numpy(x)
    t_y = torch.from_numpy(y)
    x1 = trans_to_numpy(t_x[3])
    # image = io.imshow(x1)

    # Max池化层,conv2d3kernel  输出通道,
    cfgs = {
        'A': [64, 'M', 128, 256],
    }


    def vgg_self(cfg="A", batch_norm=False):

        model = VGG(make_layers(cfgs[cfg], batch_norm=batch_norm), num_classes=100)

        return model


    model = vgg_self(cfg="A", batch_norm=False)
    target_layer = model.features[-1]
    opitmizer = torch.optim.SGD(model.parameters(), lr=0.03)
    loss_fun = nn.MSELoss()
    for i in range(10):
        predict = model(t_x)
        # print(predict)
        loss = loss_fun(predict, t_y)
        loss = torch.as_tensor(loss, dtype=torch.float)

        print(float(loss))

        opitmizer.zero_grad()
        loss.backward()
        opitmizer.step()
    # # In [7]:
    # # # the target layer you want to visualize
    # target_layer = model.features[-1].conv2
    # # #
    # # # wrapper for class activation mapping. Choose one of the following.
    # wrapped_model = CAM(model, target_layer)
    wrapped_model = GradCRM(model, target_layer)
    # wrapped_model = GradCAMpp(model, target_layer)
    # wrapped_model = SmoothGradCAMpp(model, target_layer, n_samples=5, stdev_spread=0.15)
    # In [8]:
    tensor = t_x[5].unsqueeze(0)
    cam, idx = wrapped_model(tensor, idx=5)

    # # visualize only cam
    imshow(cam.squeeze().numpy(), alpha=0.5, cmap='jet')

    # In [11]:
    # # reverse normalization for display
    img = normalize.inverse_transform(tensor)
    # In [12]:
    heatmap = visualize(img, cam)
    # In [13]:
    # # save image
    ##save_image(heatmap, './sample/{}_cam.png'.format(idx2label[idx]).replace(" ", "_").replace(",", ""))
    # # save_image(heatmap, './sample/{}_gradcam.png'.format(idx2label[idx]).replace(" ", "_").replace(",", ""))
    # # save_image(heatmap, './sample/{}_gradcampp.png'.format(idx2label[idx]).replace(" ", "_").replace(",", ""))
    # save_image(heatmap, './sample/{}_smoothgradcampp.png'.format(idx2label[idx]).replace(" ", "_").replace(",", ""))
    # In [14]:
    # # or visualize on Jupyter
    hm = (heatmap.squeeze().numpy().transpose(1, 2, 0))
    imshow(hm)



# -*- coding: utf-8 -*-

# @Time   : 2019/5/25 17:28
# @Author : Administrator
# @Project : feature_preparation
# @FileName: exhaustion.py
# @Software: PyCharm

"""
the estimator_ is the model with the best feature rather than all feature combination
"""

import warnings
from functools import partial
from itertools import combinations

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.base import MetaEstimatorMixin
from sklearn.base import clone
from sklearn.feature_selection._base import SelectorMixin
from sklearn.utils.metaestimators import if_delegate_has_method
from sklearn.utils.validation import check_is_fitted, check_X_y

from featurebox.tools.tool import parallelize
from .mutibase import MutiBase

warnings.filterwarnings("ignore")


class Exhaustion(BaseEstimator, MetaEstimatorMixin, SelectorMixin, MutiBase):
    """
    the estimator_ is the model with the best feature rather than all feature combination
    """

    def __init__(self, estimator, n_select=(2, 3, 4), muti_grade=2, muti_index=None, must_index=None, n_jobs=1,
                 refit=False):
        """

        :param estimator: and sklearn model or GridSearchCV
        :param n_select: the n_select list,default,n_select=(3, 4)
        :param muti_grade:muti_grade=2
        :param muti_index:the range of muti_grade
        :param must_index:the columns force to index
        :param n_jobs:default=1
        """
        super().__init__(muti_grade=muti_grade, muti_index=muti_index, must_index=must_index)
        self.estimator = estimator
        self.score_ = []
        self.n_jobs = n_jobs
        self.n_select = [n_select, ] if isinstance(n_select, int) else n_select
        self.refit = refit

    @property
    def _estimator_type(self):
        return self.estimator._estimator_type

    def fit(self, X, y):
        """Fit the baf model and then the underlying estimator on the selected
           feature.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_feature]
            The training input0 samples.

        y : array-like, shape = [n_samples]
            The target values.
        """
        return self._fit(X, y)

    def _fit(self, x, y):

        def score_pri(slices, x0, y0):
            slices = list(slices)
            if len(slices) < 1:
                score0 = - np.inf
            else:
                slices = self.feature_unfold(slices)
                data_x0 = x0[:, slices]

                self.estimator.fit(data_x0, y0)

                score0 = np.mean(self.estimator.best_score_)  # score_test

                # print(slices, score0)
            return score0

        score = partial(score_pri, x0=x, y0=y)

        self.score_ = []
        x, y = check_X_y(x, y, "csc")
        assert all((self.check_must, self.check_muti)) in [True, False]

        feature_list = list(range(x.shape[1]))
        fold_feature_list = self.feature_fold(feature_list)
        if self.check_must:
            fold_feature_list = [i for i in fold_feature_list if i not in self.check_must]

        slice_all = [combinations(fold_feature_list, i) for i in self.n_select]
        slice_all = [list(self.feature_must_fold(_)) for i in slice_all for _ in i]

        scores = parallelize(n_jobs=self.n_jobs, func=score, iterable=slice_all)

        feature_combination = [self.feature_unfold(_) for _ in slice_all]
        index = np.argmax(scores)
        select_feature = feature_combination[index]
        su = np.zeros(x.shape[1], dtype=np.bool)
        su[select_feature] = 1
        self.best_score_ = max(scores)
        self.score_ = scores
        self.support_ = su
        self.estimator_ = clone(self.estimator)
        if self.refit:
            self.estimator_.fit(x[:, select_feature], y)
        self.n_feature_ = len(select_feature)
        self.score_ex = list(zip(feature_combination, scores))
        self.scatter = list(zip([len(i) for i in slice_all], scores))
        self.score_ex.sort(key=lambda _: _[1], reverse=True)

        return self

    @if_delegate_has_method(delegate='estimator')
    def predict(self, X):
        """Reduce X to the selected feature and then Fit using the
           underlying estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        Returns
        -------
        y : array of shape [n_samples]
            The predicted target values.
        """
        check_is_fitted(self, 'estimator_')
        return self.estimator_.predict(self.transform(X))

    @if_delegate_has_method(delegate='estimator')
    def score(self, X, y):
        """Reduce X to the selected feature and then return the score of the
           underlying estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_feature]
            The input0 samples.

        y : array of shape [n_samples]
            The target values.
        """
        check_is_fitted(self, 'estimator_')
        return self.estimator_.score(self.transform(X), y)

    def _get_support_mask(self):
        check_is_fitted(self, 'support_')
        return self.support_



#!/usr/bin/python3.7
# -*- coding: utf-8 -*-

# @Time   : 2019/7/29 19:48
# @Author : Administrator
# @Software: PyCharm
# @License: BSD 3-Clause


"""
# Just a copy from xenonpy

"""

import os
from os import remove

import joblib
import numpy as np
import pandas as pd
from skimage import io


def def_pwd(path):
    if path is None:
        path = os.getcwd()

    if os.path.exists(path):
        os.chdir(path)
    else:
        os.makedirs(path)
        os.chdir(path)
    pwd = os.getcwd()
    locals()[pwd] = pwd
    return pwd


class Store(object):
    """
    Store file to path.
    default model is "w" ,which can be"a+" in txt.
    """

    def __init__(self, path=None, filename="filename", prefix: str = None):
        r"""
        store file in path
        :param filename: str,universal filename
        :param path: None, ..
        \data_cluster, or F:data_cluster\data1
        """

        if not prefix:
            prefix = ""
        self._prefix = prefix

        def_pwd(path)

        self._path = path
        self._filename = ""
        self.default_filename = filename
        self._file_list = []

    def __repr__(self):
        return "store to ({}) with {} file".format(self._path, len(self.stored_file))

    __str__ = __repr__

    def _check_name(self, suffix="csv", file_new_name="filename", model="w"):

        self._filename = file_new_name or self.default_filename

        if os.path.isfile('{}{}.{}'.format(self._prefix, self._filename, suffix)) and model == "w":
            shu1 = 1
            while os.path.isfile('{}{}({}).{}'.format(self._prefix, self._filename, shu1, suffix)):
                shu1 += 1
            self._filename = '{}{}({}).{}'.format(self._prefix, self._filename, shu1, suffix)
        else:
            self._filename = '{}{}.{}'.format(self._prefix, self._filename, suffix)

        if self._filename in self._file_list:
            self._file_list.remove(self._filename)
        self._file_list.append(self._filename)

    def to_csv(self, data, file_new_name=None, model="w", reverse=False):

        self._check_name("csv", file_new_name, model=model)
        if isinstance(data, (dict, list)):
            data = pd.DataFrame.from_dict(data).T

        elif isinstance(data, np.ndarray):
            data = pd.DataFrame(data)
        if isinstance(data, pd.DataFrame):
            if reverse:
                data = data.T
            data.to_csv(path_or_buf="%s" % self._filename, sep=",", na_rep='', float_format=None,
                        columns=None, header=True, index=True, index_label=None,
                        mode=model, encoding=None, )

        else:
            raise TypeError("Not support data_cluster type:%s for csv" % type(data))

    def to_txt(self, data, file_new_name=None, model="w"):
        self._check_name("txt", file_new_name, model=model)
        document = open(self._filename, model)
        document.write(str(data))
        document.close()

    def to_pkl_pd(self, data, file_new_name=None):
        self._check_name("pkl_pd", file_new_name)
        pd.to_pickle(data, self._filename)

    def to_pkl_sk(self, data, file_new_name=None):
        self._check_name("pkl_sk", file_new_name)
        joblib.dump(data, self._filename)

    def to_png(self, data, file_new_name=None):
        self._check_name("png", file_new_name=file_new_name)
        io.imsave(self._filename, data)

    @classmethod
    def to_multi_file(cls, datas, suffix="pkl_sk", file_new_name=None):
        dict_func = {"txt": cls.to_txt, "pkl_sk": cls.to_pkl_sk, "pkl_pd": cls.to_pkl_pd,
                     "csv": cls.to_csv, "png": cls.to_png}
        for data in datas:
            dict_func[suffix](data, file_new_name, model="w")

    def remove(self, index_or_name=None):
        """
        remove the indexed file
        :param index_or_name: index or x_name,default=-1

        """
        if isinstance(index_or_name, str):
            name = index_or_name
            index = -1
        elif isinstance(index_or_name, int):
            name = None
            index = index_or_name
        else:
            name = None
            index = -1
        if not name:
            try:
                files = self._file_list[index]
            except IndexError:
                raise IndexError("No flie or wrong index to remove")
            else:
                if not isinstance(files, list):
                    remove(str(files))
                else:
                    for f in files:
                        remove(str(f))
                del self._file_list[index]

        elif name in self._file_list:
            if not isinstance(name, list):
                remove(str(name))
                self._file_list.remove(name)
            else:
                for f in name:
                    remove(str(f))
                    self._file_list.remove(f)
                self._file_list.remove([])

        else:
            raise NameError("No flie named %s" % index_or_name)

    def withdraw(self):
        """
        delet all stored_file file
        """
        files = self._file_list
        for f in files:
            remove(str(f))
        self._file_list = []
        self._filename = ""

    @property
    def stored_file(self):
        [print(i) for i in self._file_list]
        return self._file_list


if __name__ == "__main__":
    a = np.array([[1, 2], [3, 4]])
    st = Store(r'C:\Users\Administrator\Desktop')
    st.to_txt(a, file_new_name="filename", model="a+")
